"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import datetime
import os
import uuid

import pytest

from inmanta import data
from inmanta.const import Change, ResourceAction, ResourceState
from inmanta.util import get_compiler_version
from utils import get_resource


async def test_project(server, client, cli):
    # create a new project
    result = await cli.run("project", "create", "-n", "test_project")
    assert result.exit_code == 0

    projects = await client.list_projects()
    assert len(projects.result["projects"]) == 1

    project = projects.result["projects"][0]
    assert project["id"] in result.output
    assert project["name"] in result.output

    # show the project
    result = await cli.run("project", "show", project["id"])
    assert result.exit_code == 0
    assert project["id"] in result.output
    assert project["name"] in result.output

    result = await cli.run("project", "show", project["name"])
    assert result.exit_code == 0
    assert project["id"] in result.output
    assert project["name"] in result.output

    # modify the project
    new_name = "test_project_2"
    result = await cli.run("project", "modify", "-n", new_name, project["name"])
    assert result.exit_code == 0
    assert project["id"] in result.output
    assert new_name in result.output

    new_name = "test_project_3"
    result = await cli.run("project", "modify", "-n", new_name, project["id"])
    assert result.exit_code == 0
    assert project["id"] in result.output
    assert new_name in result.output

    # delete the project
    result = await cli.run("project", "delete", project["id"])
    assert result.exit_code == 0


async def test_environment(server, client, cli, tmpdir):
    project_name = "test_project"
    result = await client.create_project(project_name)
    assert result.code == 200
    project_id = result.result["project"]["id"]

    # create a new environment
    result = await cli.run("environment", "create", "-n", "test1", "-r", "/git/repo", "-b", "dev1", "-p", project_name)
    assert result.exit_code == 0

    result = await cli.run("environment", "create", "-n", "test2", "-r", "/git/repo", "-b", "dev2", "-p", project_id)
    assert result.exit_code == 0

    environments = await client.list_environments()
    assert len(environments.result["environments"]) == 2
    environments = environments.result["environments"]

    # list environments
    result = await cli.run("environment", "list")
    assert result.exit_code == 0
    assert "test_project" in result.output
    assert "test1" in result.output
    assert "test2" in result.output

    # show an environment
    env_name = environments[0]["name"]
    env_id = environments[0]["id"]

    result = await cli.run("environment", "show", env_name)
    assert result.exit_code == 0
    assert env_name in result.output
    assert env_id in result.output

    result = await cli.run("environment", "show", "--format", "{id}--{name}--{project}--{repo_url}--{repo_branch}", env_name)
    assert result.exit_code == 0
    assert result.output.strip() == f"{env_id}--{env_name}--{project_id}--/git/repo--dev1"

    result = await cli.run("environment", "show", env_id)
    assert result.exit_code == 0
    assert env_name in result.output
    assert env_id in result.output

    os.chdir(tmpdir)
    result = await cli.run("environment", "save", env_name)
    assert result.exit_code == 0

    path_dot_inmanta_file = os.path.join(tmpdir, ".inmanta")
    assert os.path.isfile(path_dot_inmanta_file)
    with open(path_dot_inmanta_file, encoding="utf-8") as f:
        file_content = f.read()
        assert f"environment={env_id}" in file_content


async def test_environment_recompile(server, environment, client, cli):
    result = await client.get_compile_reports(environment)
    assert result.code == 200
    assert len(result.result["data"]) == 0

    result = await cli.run("environment", "recompile", environment)
    assert result.exit_code == 0
    assert "Recompile triggered successfully" in result.output

    result = await cli.run("environment", "recompile", environment, "--update")
    assert result.exit_code == 0
    assert "Update & Recompile triggered successfully" in result.output

    result = await client.get_compile_reports(environment)
    assert result.code == 200
    assert len(result.result["data"]) == 2

    await client.set_setting(environment, data.SERVER_COMPILE, False)
    result = await cli.run("environment", "recompile", environment)
    assert result.exit_code == 0
    assert "Skipping compile" in result.output
    assert "Recompile triggered successfully" not in result.output


async def test_environment_settings(server, environment, client, cli):
    result = await cli.run("environment", "setting", "list", "-e", environment)
    assert result.exit_code == 0

    result = await cli.run("environment", "setting", "set", "-e", environment, "-k", "auto_deploy", "-o", "true")
    assert result.exit_code == 0
    result = await cli.run("environment", "setting", "set", "-e", environment, "--key", "auto_deploy", "--value", "true")
    assert result.exit_code == 0

    result = await cli.run("environment", "setting", "list", "-e", environment)
    assert result.exit_code == 0
    assert environment in result.output
    assert "auto_deploy" in result.output

    result = await cli.run("environment", "setting", "get", "-e", environment, "--key", "auto_deploy")
    assert result.exit_code == 0
    assert "True" in result.output

    result = await cli.run("environment", "setting", "delete", "-e", environment, "--key", "auto_deploy")
    assert result.exit_code == 0


async def test_agent(server, client, environment, cli):
    result = await cli.run("agent", "list", "-e", environment)
    assert result.exit_code == 0


@pytest.mark.parametrize("auto_start_agent", [True])  # Allow autostart
async def test_version(server, client, clienthelper, environment, cli):
    version = str(await clienthelper.get_version())
    resources = [
        {
            "key": "key1",
            "value": "value1",
            "id": "test::Resource[agent1,key=key1],v=" + version,
            "send_event": False,
            "purged": False,
            "requires": ["test::Resource[agent1,key=key2],v=" + version],
        },
        {
            "key": "key2",
            "value": "value2",
            "id": "test::Resource[agent1,key=key2],v=" + version,
            "send_event": False,
            "requires": [],
            "purged": False,
        },
        {
            "key": "key3",
            "value": None,
            "id": "test::Resource[agent1,key=key3],v=" + version,
            "send_event": False,
            "requires": [],
            "purged": True,
        },
    ]

    result = await client.put_version(
        tid=environment,
        version=version,
        resources=resources,
        unknowns=[],
        version_info={},
        compiler_version=get_compiler_version(),
    )
    assert result.code == 200

    result = await cli.run("version", "list", "-e", environment)
    assert result.exit_code == 0
    assert version in result.output
    assert "candidate" in result.output

    result = await cli.run("version", "release", "-e", environment, version)
    assert result.exit_code == 0
    assert version in result.output

    result = await client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["released"]

    result = await cli.run("version", "report", "-e", environment, "-i", version)
    assert result.exit_code == 0


async def test_param(server, client, environment, cli):
    await client.set_setting(environment, data.SERVER_COMPILE, False)

    result = await cli.run("param", "set", "-e", environment, "--name", "var1", "--value", "value1")
    assert result.exit_code == 0
    assert "value1" in result.output

    result = await cli.run("param", "get", "-e", environment, "--name", "var1")
    assert result.exit_code == 0
    assert "value1" in result.output

    result = await cli.run("param", "list", "-e", environment)
    assert result.exit_code == 0
    assert "var1" in result.output


async def test_create_environment(tmpdir, server, client, cli):
    """
    Tests the "inmanta-cli environment create" command
    and overwrite prompt/overwriting of the .inmanta file
    """
    os.chdir(tmpdir)
    file_path = os.path.join(os.getcwd(), ".inmanta")
    result = await client.create_project("test")
    assert result.code == 200

    result = await cli.run("environment", "create", "-n", "test-env-0", "-p", "test", "--save")
    assert result.exit_code == 0
    assert os.path.exists(file_path)

    with open(file_path) as inmanta_file:
        file_content_0 = inmanta_file.read()
    ctime_0 = os.path.getctime(file_path)

    result = await cli.run("environment", "create", "-n", "test-env-1", "-p", "test", "--save", input="n")
    assert result.exit_code == 0

    with open(file_path) as inmanta_file:
        file_content_1 = inmanta_file.read()

    ctime_1 = os.path.getctime(file_path)

    assert file_content_0 == file_content_1
    assert ctime_0 == ctime_1

    result = await cli.run("environment", "create", "-n", "test-env-2", "-p", "test", "--save", input="y")
    assert result.exit_code == 0

    with open(file_path) as inmanta_file:
        file_content_2 = inmanta_file.read()

    ctime_2 = os.path.getctime(file_path)

    assert file_content_0 != file_content_2
    assert ctime_0 != ctime_2


async def test_inmanta_cli_http_version(server, client, cli):
    result = await cli.run("project", "create", "-n", "test_project")
    assert result.exit_code == 0


async def test_pause_agent(server, cli):
    project = data.Project(name="test")
    await project.insert()
    env1 = data.Environment(name="env1", project=project.id)
    await env1.insert()
    env2 = data.Environment(name="env2", project=project.id)
    await env2.insert()

    await data.Agent(environment=env1.id, name="agent1", paused=False).insert()
    await data.Agent(environment=env1.id, name="agent2", paused=False).insert()
    await data.Agent(environment=env2.id, name="agent3", paused=False).insert()

    async def assert_agent_paused(env_id: uuid.UUID, expected_records: dict[str, bool]) -> None:
        result = await cli.run("agent", "list", "-e", str(env_id))
        assert result.exit_code == 0
        output = result.stdout.replace(" ", "")
        assert "Agent|Environment|Paused" in output
        for agent_name, paused in expected_records.items():
            assert f"{agent_name}|{env_id}|{paused}" in output

    await assert_agent_paused(env_id=env1.id, expected_records=dict(agent1=False, agent2=False))
    await assert_agent_paused(env_id=env2.id, expected_records=dict(agent3=False))

    # Pause
    result = await cli.run("agent", "pause", "-e", str(env1.id), "--agent", "agent1")
    assert result.exit_code == 0
    await assert_agent_paused(env_id=env1.id, expected_records=dict(agent1=True, agent2=False))
    await assert_agent_paused(env_id=env2.id, expected_records=dict(agent3=False))

    # Unpause
    result = await cli.run("agent", "unpause", "-e", str(env1.id), "--agent", "agent1")
    assert result.exit_code == 0
    await assert_agent_paused(env_id=env1.id, expected_records=dict(agent1=False, agent2=False))
    await assert_agent_paused(env_id=env2.id, expected_records=dict(agent3=False))

    # Pause all agents in env1
    result = await cli.run("agent", "pause", "-e", str(env1.id), "--all")
    assert result.exit_code == 0
    await assert_agent_paused(env_id=env1.id, expected_records=dict(agent1=True, agent2=True))
    await assert_agent_paused(env_id=env2.id, expected_records=dict(agent3=False))

    # Unpause all agents in env1
    result = await cli.run("agent", "unpause", "-e", str(env1.id), "--all")
    assert result.exit_code == 0
    await assert_agent_paused(env_id=env1.id, expected_records=dict(agent1=False, agent2=False))
    await assert_agent_paused(env_id=env2.id, expected_records=dict(agent3=False))

    # Mandatory option -e not specified
    for action in ["pause", "unpause"]:
        result = await cli.run("agent", action, "--agent", "agent1")
        assert result.exit_code != 0

    # --agent and --all are both set
    for action in ["pause", "unpause"]:
        result = await cli.run("agent", action, "-e", str(env1.id), "--agent", "agent1", "--all")
        assert result.exit_code != 0

    # --agent and --all are both not set
    for action in ["pause", "unpause"]:
        result = await cli.run("agent", action, "-e", str(env1.id))
        assert result.exit_code != 0


async def test_list_actionlog(server, environment, client, cli, null_agent, clienthelper):
    """
    Test the `inmanta-cli action-log list` command.
    """
    await clienthelper.set_auto_deploy(auto=True)

    def assert_nr_records_in_output_table(output: str, nr_records: int) -> None:
        lines = [line.strip() for line in output.split("\n") if line.strip() and line.strip().startswith("|")]
        actual_nr_of_records = len(lines) - 1  # Exclude the header
        assert nr_records == actual_nr_of_records

    result = await client.reserve_version(tid=environment)
    assert result.code == 200
    version = result.result["data"]

    resource1 = get_resource(version, key="test1")
    resource2 = get_resource(version, key="test2")
    await clienthelper.put_version_simple(resources=[resource1, resource2], version=version, wait_for_released=True)

    result = await null_agent._client.resource_action_update(
        tid=environment,
        resource_ids=[resource1["id"]],
        action_id=uuid.uuid4(),
        action=ResourceAction.deploy,
        started=datetime.datetime.now(),
        finished=datetime.datetime.now(),
        status=ResourceState.failed,
        messages=[
            {
                "level": "INFO",
                "msg": "Deploying",
                "timestamp": datetime.datetime.now().isoformat(timespec="microseconds"),
                "args": [],
            },
            {
                "level": "ERROR",
                "msg": "Deployment failed",
                "timestamp": datetime.datetime.now().isoformat(timespec="microseconds"),
                "args": [],
                "status": ResourceState.failed.value,
            },
        ],
        changes={},
        change=Change.nochange,
        send_events=False,
    )
    assert result.code == 200, result.result
    result = await null_agent._client.resource_action_update(
        tid=environment,
        resource_ids=[resource2["id"]],
        action_id=uuid.uuid4(),
        action=ResourceAction.deploy,
        started=datetime.datetime.now(),
        finished=datetime.datetime.now(),
        status=ResourceState.deployed,
        messages=[
            {
                "level": "INFO",
                "msg": "Deployed successfully",
                "timestamp": datetime.datetime.now().isoformat(timespec="microseconds"),
                "args": [],
            },
        ],
        changes={},
        change=Change.nochange,
        send_events=False,
    )
    assert result.code == 200

    # Get all resource actions for resource1
    result = await cli.run("action-log", "list", "-e", str(environment), "--rvid", resource1["id"])
    assert result.exit_code == 0
    assert_nr_records_in_output_table(result.output, nr_records=2)  # 1 store action + 1 deploy action

    # Get deploy resource actions for resource1
    result = await cli.run("action-log", "list", "-e", str(environment), "--rvid", resource1["id"], "--action", "deploy")
    assert result.exit_code == 0
    assert_nr_records_in_output_table(result.output, nr_records=1)  # 1 deploy action

    # Resource id is provided instead of resource version id
    resource_id = resource1["id"].rsplit(",", maxsplit=1)[0]
    result = await cli.run("action-log", "list", "-e", str(environment), "--rvid", resource_id)
    assert result.exit_code != 0
    assert f"Invalid value for '--rvid': {resource_id}" in result.stderr

    # Incorrect format resource version id
    result = await cli.run("action-log", "list", "-e", str(environment), "--rvid", "test")
    assert result.exit_code != 0
    assert "Invalid value for '--rvid': test" in result.stderr


async def test_show_messages_actionlog(server, environment, client, cli, null_agent, clienthelper):
    """
    Test the `inmanta-cli action-log show-messages` command.
    """
    await clienthelper.set_auto_deploy(auto=True)

    result = await client.reserve_version(tid=environment)
    assert result.code == 200
    version = result.result["data"]

    resource1 = get_resource(version, key="test1")
    await clienthelper.put_version_simple(resources=[resource1], version=version, wait_for_released=True)

    result = await null_agent._client.resource_action_update(
        tid=environment,
        resource_ids=[resource1["id"]],
        action_id=uuid.uuid4(),
        action=ResourceAction.deploy,
        started=datetime.datetime.now(),
        finished=datetime.datetime.now(),
        status=ResourceState.deployed,
        messages=[
            {
                "level": "DEBUG",
                "msg": "Started deployment",
                "timestamp": datetime.datetime.now().isoformat(timespec="microseconds"),
                "args": [],
            },
            {
                "level": "INFO",
                "msg": "Deployed successfully",
                "timestamp": datetime.datetime.now().isoformat(timespec="microseconds"),
                "args": [],
            },
        ],
        changes={},
        change=Change.nochange,
        send_events=False,
    )
    assert result.code == 200

    # Obtain action_id
    result = await client.get_resource(tid=environment, id=resource1["id"], logs=True, log_action=ResourceAction.deploy)
    assert result.code == 200
    assert len(result.result["logs"]) == 1
    action_id = result.result["logs"][0]["action_id"]

    result = await cli.run(
        "action-log", "show-messages", "-e", str(environment), "--rvid", resource1["id"], "--action-id", str(action_id)
    )
    assert result.exit_code == 0
    assert "DEBUG Started deployment" in result.output
    assert "INFO Deployed successfully" in result.output


async def test_monitor(server, environment, client, cli, agent, clienthelper, resource_container):
    """
    Test the `inmanta-cli monitor` command.
    """
    result = await client.reserve_version(tid=environment)
    assert result.code == 200
    version = result.result["data"]

    resource1 = get_resource(version, key="test1")
    await clienthelper.set_auto_deploy()
    await clienthelper.put_version_simple(resources=[resource1], version=version, wait_for_released=True)

    result = await cli.run("monitor", "-e", environment)
    assert result.exit_code == 0



"""
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import os.path
import shutil
import uuid
from configparser import ConfigParser
from typing import Iterator

import pytest

from inmanta import config
from inmanta.config import Config
from inmanta.server import SLICE_SERVER
from utils import wait_for_version


@pytest.fixture
def compiler_logging_config(tmpdir, monkeypatch):
    """
    Set up the logging config for the compiler to use a specific formatter.
    """
    compiler_logging_config = """
        disable_existing_loggers: false
        formatters:
          console_formatter:
            format: "COMPILER_CONFIG_FLAG -- %(message)s"
        handlers:
          console_handler:
            class: logging.StreamHandler
            formatter: console_formatter
            level: DEBUG
            stream: ext://sys.stdout
        root:
          handlers:
          - console_handler
          level: INFO
        version: 1
    """
    compiler_logging_config_file = os.path.join(tmpdir, "compiler.yml")
    with open(compiler_logging_config_file, "w") as fh:
        fh.write(compiler_logging_config)
    config = os.path.join(tmpdir, "logging_config.yml")
    with open(config, "w") as fh:
        fh.write(
            f"""
[logging]
compiler = {os.path.abspath(compiler_logging_config_file)}
        """
        )
    return config


@pytest.fixture(scope="function")
def inmanta_config(clean_reset, compiler_logging_config) -> Iterator[ConfigParser]:
    Config.load_config(min_c_config_file=compiler_logging_config)
    yield config.Config.get_instance()


@pytest.mark.slowtest
async def test_server_passing_compiler_logging_config(server, client, environment):
    """
    Test that the server passes down the logging config to the compiler when starting it.
    """

    project_dir = os.path.join(server.get_slice(SLICE_SERVER)._server_storage["server"], str(environment), "compiler")
    project_source = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "project")

    shutil.copytree(project_source, project_dir)

    # add main.cf
    with open(os.path.join(project_dir, "main.cf"), "w", encoding="utf-8") as fd:
        fd.write(
            """
        import std::testing

        host = std::Host(name="test", os=std::linux)
        std::testing::NullResource(name=host.name)
    """
        )

    result = await client.notify_change(environment)
    assert result.code == 200

    versions = await wait_for_version(client, environment, 1, compile_timeout=40)
    assert versions["versions"][0]["total"] == 1
    assert versions["versions"][0]["version_info"]["export_metadata"]["type"] == "api"

    reports = await client.get_reports(environment)
    assert reports.code == 200
    assert len(reports.result["reports"]) == 1
    compile_id = reports.result["reports"][0]["id"]

    report = await client.get_report(uuid.UUID(compile_id))
    assert report.code == 200

    # Get the compile outstream
    for report in report.result["report"]["reports"]:
        if report["name"] == "Recompiling configuration model":
            compile_outstream = report["outstream"]
            assert "COMPILER_CONFIG_FLAG -- Starting compile" in compile_outstream
            break
    else:
        assert False, "Compile report doesn't contain a 'Recompiling configuration model' section."



"""
Copyright 2019 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com

Module defining the v1 rest api
"""

import datetime
import uuid
from collections.abc import Mapping, Sequence
from typing import Any, Literal, Optional, Union

import inmanta.types
from inmanta import const, data, resources
from inmanta.const import ResourceState
from inmanta.data import model
from inmanta.data.model import PipConfig
from inmanta.protocol import exceptions
from inmanta.protocol.common import ArgOption
from inmanta.protocol.decorators import method, typedmethod
from inmanta.types import JsonType, PrimitiveTypes


async def convert_environment(env: uuid.UUID, metadata: dict) -> "data.Environment":
    """

    :meta private:
    """
    metadata[const.INMANTA_URN + "env"] = str(env)
    env = await data.Environment.get_by_id(env)
    if env is None:
        raise exceptions.NotFound("The given environment id does not exist!")
    return env


async def add_env(env: uuid.UUID, metadata: dict) -> uuid.UUID:
    """

    :meta private:
    """
    metadata[const.INMANTA_URN + "env"] = str(env)
    return env


async def ignore_env(obj: Any, metadata: dict) -> Any:
    """
    This mapper only adds an env all for authz

    :meta private:
    """
    metadata[const.INMANTA_URN + "env"] = "all"
    return obj


async def convert_resource_version_id(rvid: inmanta.types.ResourceVersionIdStr, metadata: dict) -> "resources.Id":
    """

    :meta private:
    """
    try:
        return resources.Id.parse_resource_version_id(rvid)
    except Exception:
        raise exceptions.BadRequest(f"Invalid resource version id: {rvid}")


ENV_OPTS: dict[str, ArgOption] = {
    "tid": ArgOption(header=const.INMANTA_MT_HEADER, reply_header=True, getter=convert_environment)
}
AGENT_ENV_OPTS: dict[str, ArgOption] = {"tid": ArgOption(header=const.INMANTA_MT_HEADER, reply_header=True, getter=add_env)}
RVID_OPTS = {"rvid": ArgOption(getter=convert_resource_version_id)}


# Method for working with projects
@method(path="/project", operation="PUT", client_types=[const.ClientType.api])
def create_project(name: str, project_id: Optional[uuid.UUID] = None):
    """
    Create a new project

    :param name: The name of the project
    :param project_id: Optional. A unique uuid, when it is not provided the server generates one
    """


@method(path="/project/<id>", operation="POST", client_types=[const.ClientType.api])
def modify_project(id: uuid.UUID, name: str):
    """
    Modify the given project.

    :param id: The id of the project to modify.
    :param name: The new name for the project.
    """


@method(path="/project/<id>", operation="DELETE", client_types=[const.ClientType.api])
def delete_project(id: uuid.UUID):
    """
    Delete the given project and all related data.

    :param id: The id of the project to be deleted.
    """


@method(path="/project", operation="GET", client_types=[const.ClientType.api])
def list_projects():
    """
    Returns a list of projects ordered alphabetically by name. The environments within each project are also sorted by name.
    """


@method(path="/project/<id>", operation="GET", client_types=[const.ClientType.api])
def get_project(id: uuid.UUID):
    """
    Get a project and a list of the ids of all environments.

    :param id: The id of the project to retrieve.
    """


# Methods for working with environments
@method(path="/environment", operation="PUT", client_types=[const.ClientType.api])
def create_environment(
    project_id: uuid.UUID,
    name: str,
    repository: Optional[str] = None,
    branch: Optional[str] = None,
    environment_id: Optional[uuid.UUID] = None,
):
    """
    Create a new environment

    :param project_id: The id of the project this environment belongs to
    :param name: The name of the environment.
    :param repository: Optional. The URL of the repository.
    :param branch: Optional. The name of the branch in the repository.
    :param environment_id: Optional. A unique environment id, if none an id is allocated by the server.
    """


@method(path="/environment/<id>", operation="POST", client_types=[const.ClientType.api])
def modify_environment(id: uuid.UUID, name: str, repository: Optional[str] = None, branch: Optional[str] = None):
    """
    Modify the given environment.

    :param id: The id of the environment to modify.
    :param name: The new name for the environment.
    :param repository: Optional. The URL of the repository.
    :param branch: Optional. The name of the branch in the repository.

    If 'repository' or 'branch' is provided as None, the corresponding attribute of the environment remains unchanged.
    """


@method(path="/environment/<id>", operation="DELETE", client_types=[const.ClientType.api])
def delete_environment(id: uuid.UUID):
    """
    Delete the given environment and all related data.

    :param id: The id of the environment to be deleted.

    :raises NotFound: The given environment doesn't exist.
    :raises Forbidden: The given environment is protected.
    """


@method(path="/environment", operation="GET", client_types=[const.ClientType.api])
def list_environments():
    """
    Returns a list of environments. The results are sorted by (project id, environment name, environment id).
    """


@method(
    path="/environment/<id>",
    operation="GET",
    client_types=[const.ClientType.api],
    arg_options={"id": ArgOption(getter=add_env)},
)
def get_environment(id: uuid.UUID, versions: Optional[int] = None, resources: Optional[int] = None):
    """
    [DEPRECATED] Deprecated in favor of the v2 endpoint.
    Get an environment and all versions associated.

    :param id: The id of the environment to return.
    :param versions: Optional. If provided and greater than 0, include this many of the most recent versions for this
                     environment, ordered in descending order of their version number.
                     If not provided or 0, no version information is included.
    :param resources: Optional. If provided and greater than 0, include a summary of the resources in the environment.
    """


# Method for listing/getting/setting/removing settings of an environment. This API is also used by agents to configure
# environments.


@method(
    path="/environment_settings",
    operation="GET",
    arg_options=ENV_OPTS,
    api=True,
    agent_server=True,
    client_types=[const.ClientType.api, const.ClientType.agent, const.ClientType.compiler],
)
def list_settings(tid: uuid.UUID):
    """
    List the settings in the current environment ordered by name alphabetically.

    :param tid: The id of the environment to list settings for.
    """


@method(
    path="/environment_settings/<id>",
    operation="POST",
    arg_options=ENV_OPTS,
    api=True,
    agent_server=True,
    client_types=[const.ClientType.api, const.ClientType.agent, const.ClientType.compiler],
)
def set_setting(tid: uuid.UUID, id: str, value: Union[PrimitiveTypes, JsonType]):
    """
    Set a value for a setting.

    :param tid: The id of the environment.
    :param id: The id of the setting to set.
    :param value: The value to set for the setting.
    """


@method(
    path="/environment_settings/<id>",
    operation="GET",
    arg_options=ENV_OPTS,
    api=True,
    agent_server=True,
    client_types=[const.ClientType.api, const.ClientType.agent],
)
def get_setting(tid: uuid.UUID, id: str):
    """
    Get the value of a setting.

    :param tid: The id of the environment.
    :param id: The id of the setting to retrieve.
    """


@method(
    path="/environment_settings/<id>",
    operation="DELETE",
    arg_options=ENV_OPTS,
    api=True,
    agent_server=True,
    client_types=[const.ClientType.api, const.ClientType.agent],
)
def delete_setting(tid: uuid.UUID, id: str):
    """
    Restore the given setting to its default value.

    :param tid: The id of the environment from which the setting is to be deleted.
    :param id: The key of the setting to delete.

    """


# Method for listing and creating auth tokens for an environment that can be used by the agent and compilers


@method(
    path="/environment_auth",
    operation="POST",
    arg_options=ENV_OPTS,
    client_types=[const.ClientType.api, const.ClientType.compiler],
)
def create_token(tid: uuid.UUID, client_types: list, idempotent: bool = True):
    """
    Create or get a new token for the given client types.

    :param tid: The environment id.
    :param client_types: The client types for which this token is valid (api, agent, compiler).
    :param idempotent: Optional. The token should be idempotent, meaning it does not have an expire or issued at set,
                       so its value will not change.
    """


@method(
    path="/decommission/<id>",
    operation="DELETE",
    arg_options={"id": ArgOption(getter=convert_environment)},
    client_types=[const.ClientType.api],
)
def clear_environment(id: uuid.UUID):
    """
    Clears an environment by removing most of its associated data.
    This method deletes various components associated with the specified environment from the database,
    including agents, compile data, parameters, notifications, code, resources, and configuration models.
    However, it retains the entry in the Environment table itself and settings are kept.
    The environment will be temporarily halted during the decommissioning process.

    :param id: The id of the environment to be cleared.

    :raises NotFound: The given environment doesn't exist.
    :raises Forbidden: The given environment is protected.

    """


# Send a heartbeat to indicate that an agent is alive
@method(
    path="/heartbeat",
    operation="POST",
    agent_server=True,
    validate_sid=False,
    arg_options=ENV_OPTS,
    client_types=[const.ClientType.agent],
)
def heartbeat(sid: uuid.UUID, tid: uuid.UUID, endpoint_names: list, nodename: str, no_hang: bool = False):
    """
    Send a heartbeat to the server

    :param sid: The session ID used by this agent at this moment
    :param tid: The environment this node and its agents belongs to
    :param endpoint_names: The names of the endpoints on this node
    :param nodename: The name of the node from which the heart beat comes
    :param no_hang: Optional. don't use this call for long polling, but for connectivity check

    also registered as API method, because it is called with an invalid SID the first time
    """


@method(
    path="/heartbeat",
    operation="PUT",
    agent_server=True,
    arg_options={"sid": ArgOption(getter=ignore_env)},
    client_types=[const.ClientType.agent],
)
def heartbeat_reply(sid: uuid.UUID, reply_id: uuid.UUID, data: dict):
    """
    Send a reply back to the server

    :param sid: The session ID used by this agent at this moment
    :param reply_id: The id data is a reply to
    :param data: The data as a response to the reply
    """


# Upload, retrieve and check for file. A file is identified by a hash of its content.


@method(
    path="/file/<id>",
    operation="PUT",
    agent_server=True,
    api=True,
    client_types=[const.ClientType.api, const.ClientType.agent, const.ClientType.compiler],
    arg_options={"id": ArgOption(getter=ignore_env)},
)
def upload_file(id: str, content: str):
    """
    Upload a new file

    :param id: The id of the file
    :param content: The base64 encoded content of the file
    """


@method(
    path="/file/<id>",
    operation="HEAD",
    agent_server=True,
    api=True,
    client_types=[const.ClientType.api, const.ClientType.agent, const.ClientType.compiler],
    arg_options={"id": ArgOption(getter=ignore_env)},
)
def stat_file(id: str):
    """
    Does the file exist

    :param id: The id of the file to check
    """


@method(
    path="/file/<id>",
    operation="GET",
    agent_server=True,
    api=True,
    client_types=[const.ClientType.api, const.ClientType.agent, const.ClientType.compiler],
    arg_options={"id": ArgOption(getter=ignore_env)},
)
def get_file(id: str):
    """
    Retrieve a file

    :param id: The id of the file to retrieve
    """


@method(
    path="/file",
    api=True,
    client_types=[const.ClientType.api, const.ClientType.agent, const.ClientType.compiler],
    arg_options={"files": ArgOption(getter=ignore_env)},
)
def stat_files(files: list):
    """
    Check which files exist in the given list

    :param files: A list of file ids to check
    :return: A list of files that do not exist.
    """


# Manage resources on the server


@method(
    path="/resource/<id>",
    operation="GET",
    agent_server=True,
    validate_sid=False,
    arg_options=ENV_OPTS,
    api=True,
    client_types=[const.ClientType.api, const.ClientType.agent],
)
def get_resource(
    tid: uuid.UUID,
    id: str,
    logs: Optional[bool] = None,
    log_action: Optional[const.ResourceAction] = None,
    log_limit: int = 0,
):
    """
    Return a resource with the given id.

    :param tid: The id of the environment this resource belongs to
    :param id: Get the resource with the given resource version id
    :param logs: Optional. Include the logs in the response
    :param status: Optional. Only return the status of the resource
    :param log_action: Optional. The log action to include, leave empty/none for all actions. Valid actions are one of
                      the action strings in const.ResourceAction
    :param log_limit: Optional. Limit the number of logs included in the response, up to a maximum of 1000.
                      To retrieve more entries, use  /api/v2/resource_actions
                      (:func:`~inmanta.protocol.methods_v2.get_resource_actions`)
                      If None, a default limit (set to 1000) is applied.
    """


@method(path="/resource", operation="POST", agent_server=True, arg_options=ENV_OPTS, client_types=[const.ClientType.agent])
def resource_action_update(
    tid: uuid.UUID,
    resource_ids: list,
    action_id: uuid.UUID,
    action: const.ResourceAction,
    started: Optional[datetime.datetime] = None,
    finished: Optional[datetime.datetime] = None,
    status: Optional[Union[const.ResourceState, const.DeprecatedResourceState]] = None,
    messages: list = [],
    changes: dict = {},
    change: Optional[const.Change] = None,
    send_events: bool = False,
):
    """
    Send a resource update to the server

    :param tid: The id of the environment this resource belongs to
    :param resource_ids: The resource with the given resource_version_id id from the agent
    :param action_id: A unique id to indicate the resource action that has be updated
    :param action: The action performed
    :param started: Optional. The timestamp when this action was started. When this action (action_id) has not been saved yet,
                    started has to be defined.
    :param finished: Optional. The timestamp when this action was finished. Afterwards, no changes with the same action_id
                    can be stored. The status field also has to be set.
    :param status: Optional. The current status of the resource (if known)
    :param messages: Optional. A list of log entries to add to this entry.
    :param changes: Optional. A dict of changes to this resource. The key of this dict indicates the attributes/fields that
                   have been changed. The value contains the new value and/or the original value.
    :param change: Optional. The result of the changes
    :param send_events: Optional. [DEPRECATED] The value of this field is not used anymore.
    """


# Manage configuration model versions


@method(path="/version", operation="GET", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def list_versions(tid: uuid.UUID, start: Optional[int] = None, limit: Optional[int] = None):
    """
    Returns a list of all available versions, ordered by version number, descending

    :param tid: The id of the environment
    :param start: Optional. parameter to control the amount of results that are returned. 0 is the latest version.
    :param limit: Optional. parameter to control the amount of results returned, up to a maximum of 1000.
                  If None, a default limit (set to 1000) is applied.
    """


@method(path="/version/<id>", operation="GET", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def get_version(
    tid: uuid.UUID, id: int, include_logs: Optional[bool] = None, log_filter: Optional[str] = None, limit: Optional[int] = None
):
    """
    Get a particular version and a list of all resources in this version

    :param tid: The id of the environment
    :param id: The id of the version to retrieve
    :param include_logs: Optional. If true, a log of all operations on all resources is included
    :param log_filter: Optional. Filter log to only include actions of the specified type
    :param limit: Optional. The maximal number of actions to return per resource (starting from the latest),
                    up to a maximum of 1000.
                    To retrieve more entries, use /api/v2/resource_actions
                    (:func:`~inmanta.protocol.methods_v2.get_resource_actions`)
                    If None, a default limit (set to 1000) is applied.
    """


@method(path="/version/<id>", operation="DELETE", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def delete_version(tid: uuid.UUID, id: int):
    """
    Delete a particular version and resources

    :param tid: The id of the environment
    :param id: The id of the version to retrieve
    """


@method(path="/version", operation="PUT", arg_options=ENV_OPTS, client_types=[const.ClientType.compiler])
def put_version(
    tid: uuid.UUID,
    version: int,
    resources: list,
    resource_state: dict[inmanta.types.ResourceIdStr, Literal[ResourceState.available, ResourceState.undefined]] = {},
    unknowns: Optional[Sequence[Mapping[str, PrimitiveTypes]]] = None,
    version_info: Optional[dict] = None,
    compiler_version: Optional[str] = None,
    resource_sets: dict[inmanta.types.ResourceIdStr, Optional[str]] = {},
    pip_config: Optional[PipConfig] = None,
    project_constraints: str | None = None,
):
    """
    Store a new version of the configuration model

    The version number must be obtained through the reserve_version call

    :param tid: The id of the environment
    :param version: The version of the configuration model
    :param resources: A list of all resources in the configuration model (deployable)
    :param resource_state: A dictionary with the initial const.ResourceState per resource id. The ResourceState should be set
                           to undefined when the resource depends on an unknown or available when it doesn't.
    :param unknowns: Optional. A list of unknown parameters that caused the model to be incomplete
    :param version_info: Optional. Module version information
    :param compiler_version: Optional. version of the compiler, if not provided, this call will return an error
    :param resource_sets: Optional. a dictionary describing which resource belongs to which resource set
    :param pip_config: Optional. Pip config used by this version
    :param project_constraints: String of all the constraints set at the project level (if any) to be enforced during agent
        code install
    """


@method(
    path="/version/<id>", operation="POST", arg_options=ENV_OPTS, client_types=[const.ClientType.api, const.ClientType.compiler]
)
def release_version(
    tid: uuid.UUID, id: int, push: bool = False, agent_trigger_method: Optional[const.AgentTriggerMethod] = None
):
    """
    Release version of the configuration model for deployment.

    :param tid: The id of the environment
    :param id: The version of the CM to deploy
    :param push: [DEPRECATED] This argument is ignored.
    :param agent_trigger_method: [DEPRECATED]: This argument is ignored.

    :return: Returns the following status codes:
            200: The version is released
            404: The requested version does not exist
            409: The requested version was already released
    """


@method(path="/deploy", operation="POST", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def deploy(
    tid: uuid.UUID,
    agent_trigger_method: const.AgentTriggerMethod = const.AgentTriggerMethod.push_full_deploy,
    agents: Optional[list] = None,
):
    """
    Notify agents to perform a deploy now.

    :param tid: The id of the environment.
    :param agent_trigger_method: Indicates whether the agents should perform a full or an incremental deploy.
    :param agents: Optional, names of specific agents to trigger
    """


# Method for requesting and quering a dryrun


@method(path="/dryrun/<id>", operation="POST", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def dryrun_request(tid: uuid.UUID, id: int):
    """
    Do a dryrun

    :param tid: The id of the environment
    :param id: The version of the CM to deploy
    """


@method(path="/dryrun", operation="GET", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def dryrun_list(tid: uuid.UUID, version: Optional[int] = None):
    """
    Get the list of dry runs for an environment. The results are sorted by dry run id.

    :param tid: The id of the environment
    :param version: Optional. Only for this version
    """


@method(path="/dryrun/<id>", operation="GET", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def dryrun_report(tid: uuid.UUID, id: uuid.UUID):
    """
    Create a dryrun report

    :param tid: The id of the environment
    :param id: The version dryrun to report
    """


@method(path="/dryrun/<id>", operation="PUT", agent_server=True, arg_options=ENV_OPTS, client_types=[const.ClientType.agent])
def dryrun_update(tid: uuid.UUID, id: uuid.UUID, resource: str, changes: dict):
    """
    Store dryrun results at the server

    :param tid: The id of the environment
    :param id: The version dryrun to report
    :param resource: The id of the resource
    :param changes: The required changes
    """


# Method for requesting a dryrun from an agent


@method(
    path="/agent_dryrun/<id>",
    operation="POST",
    server_agent=True,
    timeout=5,
    arg_options=AGENT_ENV_OPTS,
    client_types=[],
    enforce_auth=False,
)
def do_dryrun(tid: uuid.UUID, id: uuid.UUID, agent: str, version: int):
    """
    Do a dryrun on an agent

    :param tid: The environment id
    :param id: The id of the dryrun
    :param agent: The agent to do the dryrun for
    :param version: The version of the model to dryrun
    """


# Method to notify the server of changes in the configuration model source code


@method(
    path="/notify/<id>",
    operation="GET",
    arg_options={"id": ArgOption(getter=convert_environment)},
    client_types=[const.ClientType.api],
)
def notify_change_get(id: uuid.UUID, update: bool = True):
    """
    Simplified GET version of the POST method

    :param id: The id of the environment.
    :param update: Optional. Indicates whether to update the model code and modules. Defaults to true.
    """


@method(
    path="/notify/<id>",
    operation="POST",
    arg_options={"id": ArgOption(getter=convert_environment)},
    client_types=[const.ClientType.api],
)
def notify_change(id: uuid.UUID, update: bool = True, metadata: dict = {}):
    """
    Notify the server that the repository of the environment with the given id, has changed.

    :param id: The id of the environment
    :param update: Optional. Update the model code and modules. Default value is true
    :param metadata: Optional. The metadata that indicates the source of the compilation trigger.
    """


@method(path="/notify/<id>", operation="HEAD", client_types=[const.ClientType.api])
def is_compiling(id: uuid.UUID):
    """
    Is a compiler running for the given environment

    :param id: The environment id
    """


# Get and set parameters on the server


@method(
    path="/parameter/<id>",
    operation="GET",
    arg_options=ENV_OPTS,
    client_types=[const.ClientType.api, const.ClientType.compiler, const.ClientType.agent],
)
def get_param(tid: uuid.UUID, id: str, resource_id: Optional[str] = None):
    """
    Get a parameter from the server.

    :param tid: The id of the environment
    :param id: The name of the parameter
    :param resource_id: Optional. scope the parameter to resource (fact),
                        if the resource id should not contain a version, the latest version is used
    :return: Returns the following status codes:

        - 200: The parameter content is returned
        - 404: The parameter is not found and unable to find it because its resource is not known to the server
        - 410: The parameter has expired
        - 503: The parameter is not found but its value is requested from an agent

    """


@method(
    path="/parameter/<id>",
    operation="PUT",
    arg_options=ENV_OPTS,
    client_types=[const.ClientType.api, const.ClientType.compiler, const.ClientType.agent],
)
def set_param(
    tid: uuid.UUID,
    id: str,
    source: const.ParameterSource,
    value: str,
    resource_id: Optional[str] = None,
    metadata: dict = {},
    recompile: bool = False,
    expires: Optional[bool] = None,
):
    """
    Set a parameter on the server. If the parameter is an tracked unknown, it will trigger a recompile on the server.
    Otherwise, if the value is changed and recompile is true, a recompile is also triggered.
    [DEPRECATED] Please use the new endpoints `/facts/<name>` and `/parameters/<name>` instead.

    :param tid: The id of the environment
    :param id: The name of the parameter
    :param source: The source of the parameter.
    :param value: The value of the parameter
    :param resource_id: Optional. Scope the parameter to resource (fact)
    :param metadata: Optional. Metadata about the parameter
    :param recompile: Optional. Whether to trigger a recompile
    :param expires: When setting a new parameter/fact: if set to None, then a sensible default will be provided (i.e. False
        for parameter and True for fact). When updating a parameter or fact, a None value will leave the existing value
        unchanged.
    """


@method(
    path="/parameter/<id>",
    operation="DELETE",
    arg_options=ENV_OPTS,
    client_types=[const.ClientType.api, const.ClientType.compiler, const.ClientType.agent],
)
def delete_param(tid: uuid.UUID, id: str, resource_id: Optional[str] = None):
    """
    Delete a parameter on the server

    :param tid: The id of the environment
    :param id: The name of the parameter
    :param resource_id: Optional. The resource id of the parameter
    """


@method(
    path="/parameter", operation="POST", arg_options=ENV_OPTS, client_types=[const.ClientType.api, const.ClientType.compiler]
)
def list_params(tid: uuid.UUID, query: dict = {}):
    """
    List/query parameters in this environment. The results are ordered alphabetically by parameter name.

    :param tid: The id of the environment
    :param query: Optional. A query to match against metadata
    """


#  Get and set parameters on the server


@method(
    path="/parameters",
    operation="PUT",
    agent_server=True,
    arg_options=ENV_OPTS,
    client_types=[const.ClientType.api, const.ClientType.compiler, const.ClientType.agent],
)
def set_parameters(tid: uuid.UUID, parameters: list):
    """
    Set a parameter on the server

    :param tid: The id of the environment
    :param parameters: A list of dicts with the following keys:

        - id The name of the parameter
        - source The source of the parameter. Valid values are defined in the ParameterSource enum (see: inmanta/const.py)
        - value The value of the parameter
        - resource_id Optionally, scope the parameter to resource (fact)
        - expires Defaults to true. Set to false to create a never expiring fact
        - metadata metadata about the parameter
    """


# Get parameters from the agent


@method(
    path="/agent_parameter",
    operation="POST",
    server_agent=True,
    timeout=5,
    arg_options=AGENT_ENV_OPTS,
    client_types=[],
    reply=False,
    enforce_auth=False,
)
def get_parameter(tid: uuid.UUID, agent: str, resource: dict):
    """
    Get all parameters/facts known by the agents for the given resource

    This method will not actually return them.
    This call wil register the request with the agent and return,
    The agent will push the parameters back to the server when they are available.

    :param tid: The environment
    :param agent: The agent to get the parameters from
    :param resource: The resource to query the parameters from
    """


@method(path="/codebatched/<id>", operation="PUT", arg_options=ENV_OPTS, client_types=[const.ClientType.compiler])
def upload_code_batched(tid: uuid.UUID, id: int, resources: dict):
    """
    Upload batches of code for various resources associated with a specific version of a configuration model in an environment.

    :param tid: The id of the environment to which the code belongs.
    :param id: The version number of the configuration model.
    :param resources: A dictionary where each key is a string representing a resource type.
                  For each resource type, the value is a dictionary. This nested dictionary's keys are file hashes,
                  and each key maps to a tuple. This tuple contains three elements: the file name, the module name,
                  and a list of requirements.

    The endpoint validates that all provided file references are valid and checks for conflicts with existing code entries.
    """


# Generate download the diff of two hashes


@method(path="/filediff", client_types=[const.ClientType.api])
def diff(file_id_1: str, file_id_2: str):
    """
    Returns the diff of the files with the two given ids

    :param file_id_1: The identifier of the first file.
    :param file_id_2: The identifier of the second file.

    :return: A string representing the diff between the two files.
    """


# Get a list of compile reports


@method(path="/compilereport", operation="GET", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def get_reports(tid: uuid.UUID, start: Optional[str] = None, end: Optional[str] = None, limit: Optional[int] = None):
    """
    Return compile reports newer then start

    The returned compile report objects may carry links to other objects, e.g. a service instance.
    The full list of supported links can be found :ref:`here <api_self_referencing_links>`.

    :param tid: The id of the environment to get a report from
    :param start: Optional. Reports after start
    :param end: Optional. Reports before end
    :param limit: Optional. Maximum number of results, up to a maximum of 1000
                  If None, a default limit (set to 1000) is applied.
    """


@method(path="/compilereport/<id>", operation="GET", client_types=[const.ClientType.api])
def get_report(id: uuid.UUID):
    """
    Get a compile report from the server

    The returned compile report object may carry links to other objects, e.g. a service instance.
    The full list of supported links can be found :ref:`here <api_self_referencing_links>`.

    :param id: The id of the compile and its reports to fetch.
    """


# Get a list of all agents


@method(path="/agentproc", operation="GET", client_types=[const.ClientType.api])
def list_agent_processes(
    environment: Optional[uuid.UUID] = None,
    expired: bool = True,
    start: Optional[uuid.UUID] = None,
    end: Optional[uuid.UUID] = None,
    limit: Optional[int] = None,
):
    """
    Return a list of all nodes and the agents for these nodes

    :param environment: Optional. An optional environment. If set, only the agents that belong to this environment are returned
    :param expired: Optional. if true show expired processes, otherwise only living processes are shown. True by default
    :param start: Optional. Agent processes after start (sorted by sid in ASC)
    :param end: Optional. Agent processes before end (sorted by sid in ASC)
    :param limit: Optional. Maximum number of results, up to a maximum of 1000
                  If None, a default limit (set to 1000) is applied.

    :raises BadRequest: limit parameter can not exceed 1000
    :raises NotFound: The given environment id does not exist!

    :return: A list of nodes
    """


@method(path="/agentproc/<id>", operation="GET", client_types=[const.ClientType.api])
def get_agent_process(id: uuid.UUID):
    """
    Return a detailed report for a node

    :param id: The session id of the agent
    :return: The requested node
    """


@method(path="/agent", operation="GET", api=True, timeout=5, arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def list_agents(tid: uuid.UUID, start: Optional[str] = None, end: Optional[str] = None, limit: Optional[int] = None):
    """
    List all agent for an environment

    [DEPRECATED] use the V2 `get_agents` or `/agents` endpoint instead

    :param tid: The environment the agents are defined in
    :param start: Optional. Agent after start (sorted by name in ASC)
    :param end: Optional. Agent before end (sorted by name in ASC)
    :param limit: Optional. Maximum number of results, up to a maximum of 1000.
                  If None, a default limit (set to 1000) is applied.

    :raises BadRequest: limit parameter can not exceed 1000
    :raises NotFound: The given environment id does not exist!
    """


# Reporting by the agent to the server


@method(path="/status", operation="GET", server_agent=True, enforce_auth=False, timeout=5, client_types=[])
def get_status():
    """
    A call from the server to the agent to report its status to the server

    :return: A map with report items
    """


# Methods to allow the server to set the agents state


@method(path="/agentstate", operation="POST", server_agent=True, enforce_auth=False, timeout=5, client_types=[])
def set_state(agent: Optional[str], enabled: bool):
    """
    Set the state of the agent.

    :param agent: The name of the agent if it's provided. None represents all agents
    :param enabled: A boolean value indicating whether the agent should be paused (enabled=False) or unpaused (enabled=True).
    """


@method(
    path="/deploy",
    operation="POST",
    server_agent=True,
    enforce_auth=False,
    timeout=5,
    arg_options=AGENT_ENV_OPTS,
    client_types=[],
)
def trigger(tid: uuid.UUID, id: None | str, incremental_deploy: bool):
    """
    When the <id> parameter is set: request this specific agent to reload resources.
    Otherwise, request ALL agents in the environment to reload resources.

    :param tid: The environment this agent is defined in
    :param id: The name of the agent
    :param incremental_deploy: Indicates whether the agent should perform an incremental deploy or a full deploy
    """


@typedmethod(
    path="/scheduler/",
    operation="POST",
    server_agent=True,
    enforce_auth=False,
    timeout=5,
    arg_options=AGENT_ENV_OPTS,
    client_types=[],
)
def trigger_read_version(tid: uuid.UUID) -> int:
    """
    Notify the scheduler that a new version has been released

    :param tid: The environment this agent is defined in
    """


# Methods for the agent to get its initial state from the server


@method(path="/agentrecovery", operation="GET", agent_server=True, arg_options=ENV_OPTS, client_types=[const.ClientType.agent])
def get_state(tid: uuid.UUID, sid: uuid.UUID, agent: str):
    """
    Get the state for this agent.

    :param tid: The id of the environment.
    :param sid: The session ID associated with this agent.
    :param agent: The name of the agent.

    :return: A map with key enabled and value a boolean.
    """


@typedmethod(path="/serverstatus", operation="GET", client_types=[const.ClientType.api])
def get_server_status() -> model.StatusResponse:
    """
    Get the status of the server
    """


@typedmethod(
    path="/compilequeue",
    operation="GET",
    arg_options=ENV_OPTS,
    client_types=[const.ClientType.api],
    api_version=1,
    envelope_key="queue",
)
def get_compile_queue(tid: uuid.UUID) -> list[model.CompileRun]:
    """
    Get the current compiler queue on the server, ordered by increasing `requested` timestamp.

    The returned compile run object may carry links to other objects, e.g. a service instance. The full list of supported links
    can be found :ref:`here <api_self_referencing_links>`.

    :param tid: The id of the environment for which to retrieve the compile queue.

    :return: A list of CompileRun objects representing the current state of the compiler queue,
             with each entry detailing a specific compile run.
    """



"""
Copyright 2021 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: bart@inmanta.com
"""

import glob
import importlib
import logging
import os
import re
import subprocess
import sys
import tempfile
from importlib.abc import Loader
from subprocess import CalledProcessError
from typing import Callable, LiteralString, Optional
from unittest.mock import patch

import py
import pytest

import inmanta.util
from inmanta import env, loader, module
from inmanta.data.model import PipConfig
from inmanta.env import Pip
from packaging import version
from utils import LogSequence, PipIndex, create_python_package


@pytest.mark.slowtest
def test_venv_pyton_env_empty_string(tmpdir, deactive_venv):
    """test that an exception is raised if the venv path is an empty string"""
    with pytest.raises(ValueError) as e:
        env.VirtualEnv("")
    assert e.value.args[0] == "The env_path cannot be an empty string."

    env_dir1 = tmpdir.mkdir("env1").strpath
    venv1 = env.VirtualEnv(env_dir1)
    venv1.use_virtual_env()

    env_dir2 = tmpdir.mkdir("env2").strpath
    venv2 = env.VirtualEnv(env_dir2)
    venv2.env_path = ""
    with pytest.raises(Exception) as e:
        venv2.use_virtual_env()
    assert e.value.args[0] == "The env_path cannot be an empty string."

    with pytest.raises(ValueError) as e:
        env.PythonEnvironment(python_path="")
    assert e.value.args[0] == "The python_path cannot be an empty string."

    with pytest.raises(ValueError) as e:
        env.PythonEnvironment(env_path="")
    assert e.value.args[0] == "The env_path cannot be an empty string."


@pytest.mark.slowtest
def test_basic_install(tmpdir, deactive_venv):
    env_dir1 = tmpdir.mkdir("env1").strpath
    venv1 = env.VirtualEnv(env_dir1)
    venv1.use_virtual_env()

    assert not venv1.are_installed(["lorem"])

    venv1.install_from_list(["lorem"])
    assert venv1.are_installed(["lorem"])

    assert not venv1.are_installed(["dummy-yummy"])

    venv1 = env.VirtualEnv(env_dir1)
    venv1.use_virtual_env()
    venv1.install_from_list(["dummy-yummy"])
    assert venv1.are_installed(["dummy-yummy"])


def test_git_based_install(tmpdir: py.path.local, deactive_venv) -> None:
    """
    Verify that the install methods can handle git-based installs over https.
    """
    venv = env.VirtualEnv(tmpdir.mkdir("env").strpath)
    venv.use_virtual_env()

    pkg_name: LiteralString = "pytest-inmanta"
    assert not venv.are_installed([pkg_name])

    try:
        venv.install_from_list([f"{pkg_name}@git+https://github.com/inmanta/{pkg_name}"])
    except CalledProcessError as ep:
        print(ep.stdout)
        raise

    assert venv.are_installed([pkg_name])


@pytest.mark.slowtest
def test_install_package_already_installed_in_parent_env(tmpdir, deactive_venv):
    """Test using and installing a package that is already present in the parent virtual environment."""
    # get all packages in the parent
    parent_installed = list(env.process_env.get_installed_packages().keys())

    # create a venv and list all packages available in the venv
    venv = env.VirtualEnv(str(tmpdir))
    venv.use_virtual_env()

    installed_packages = list(venv.get_installed_packages().keys())
    # verify that the venv sees all parent packages
    assert not set(parent_installed) - set(installed_packages)

    # site dir should be empty
    site_dir = os.path.join(venv.env_path, "lib/python*/site-packages")
    dirs = glob.glob(site_dir)
    assert len(dirs) == 1
    site_dir = dirs[0]

    def _list_dir(path: str, ignore: list[str]) -> list[str]:
        return [d for d in os.listdir(site_dir) if d not in ignore]

    # site_dir should only contain a sitecustomize.py file that sets up inheritance from the parent venv
    assert not _list_dir(site_dir, ignore=["inmanta-inherit-from-parent-venv.pth", "__pycache__"])

    # test installing a package that is already present in the parent venv
    assert "more-itertools" in parent_installed
    venv.install_from_list(["more-itertools"])

    # site_dir should only contain a sitecustomize.py file that sets up inheritance from the parent venv
    assert not _list_dir(site_dir, ignore=["inmanta-inherit-from-parent-venv.pth", "__pycache__"])

    # report json
    subprocess.check_output([os.path.join(venv.env_path, "bin/pip"), "list"])


def test_gen_req_file():
    """
    These are all examples used in older testcases that did not work correctly before
    They are supported now

    This testcase now only verifies they are all correctly parsed
    """

    reqs = [
        "lorem == 0.1.1",
        "lorem > 0.1",
        "dummy-yummy",
        "iplib@git+https://github.com/bartv/python3-iplib",
        "lorem",
        # verify support for environment markers as described in PEP 508
        "lorem;python_version<'3.7'",
        "lorem;platform_machine == 'x86_64' and platform_system == 'Linux'",
        "lorem == 0.1;python_version=='3.7'",
        "lorem == 0.2;python_version=='3.9'",
        "dep[opt]",
        "dep[otheropt]",
    ]

    # make sure they all parse
    for req in reqs:
        inmanta.util.parse_requirement(requirement=req)


def test_environment_python_version_multi_digit(tmpdir: py.path.local) -> None:
    """
    Make sure the constructor for env.Environment can handle multi-digit minor versions of Python to ensure compatibility with
    Python 3.10+.
    """
    with patch("sys.version_info", new=(3, 123, 0)):
        # python version is not included in path on windows
        with patch("sys.platform", new="linux"):
            assert env.PythonEnvironment(env_path=str(tmpdir)).site_packages_dir == os.path.join(
                str(tmpdir), "lib", "python3.123", "site-packages"
            )


@pytest.mark.slowtest
@pytest.mark.parametrize_any("version", [None, version.Version("8.6.0")])
def test_process_env_install_from_index(
    tmpdir: str,
    tmpvenv_active: tuple[py.path.local, py.path.local],
    version: Optional[version.Version],
) -> None:
    """
    Install a package from a pip index into the process_env. Assert any version specs are respected.
    """
    package_name: str = "more-itertools"
    assert package_name not in env.process_env.get_installed_packages()
    env.process_env.install_for_config(
        [inmanta.util.parse_requirement(requirement=package_name + (f"=={version}" if version is not None else ""))],
        config=PipConfig(
            use_system_config=True,  # we need an upstream for some packages
        ),
    )
    installed: dict[str, version.Version] = env.process_env.get_installed_packages()
    assert package_name in installed
    if version is not None:
        assert installed[package_name] == version
    # legacy method
    # We call it here to make sure the legacy compatibility works
    # It massages the inputs to get to the same call as the one above.
    # It should hit the cache there and return here.
    # Cheap and fast test
    env.process_env.install_from_index(
        [inmanta.util.parse_requirement(requirement=package_name + (f"=={version}" if version is not None else ""))],
        use_pip_config=True,
    )


@pytest.mark.slowtest
@pytest.mark.parametrize_any("use_extra_indexes_env", [False, True])
@pytest.mark.parametrize_any("use_extra_indexes", [False, True])
@pytest.mark.parametrize_any("use_system_config", [False, True])
def test_process_env_install_from_index_not_found_env_var(
    tmpvenv_active: tuple[py.path.local, py.path.local],
    monkeypatch,
    create_empty_local_package_index_factory: Callable[[str], str],
    use_extra_indexes: bool,
    use_extra_indexes_env: bool,
    use_system_config: bool,
) -> None:
    """
    Attempt to install a package that does not exist from the pip indexes defined in the env vars, in the pip config or in both.
    This if the system config are used or not.
    Assert the appropriate error is raised.
    """
    index_urls = [create_empty_local_package_index_factory()]

    if use_extra_indexes_env:
        extra_env_indexes = [
            create_empty_local_package_index_factory("extra_env1"),
            create_empty_local_package_index_factory("extra_env2"),
        ]
        # Convert list to a space-separated string for the environment variable
        monkeypatch.setenv("PIP_EXTRA_INDEX_URL", " ".join(extra_env_indexes))
        if use_system_config:
            # Include environment extra indexes in the main list for assertion
            index_urls.extend(extra_env_indexes)

    if use_extra_indexes:
        index_urls.extend(
            [
                create_empty_local_package_index_factory("extra1"),
                create_empty_local_package_index_factory("extra2"),
            ]
        )

    expected = (
        "Packages this-package-does-not-exist were not "
        "found in the given indexes. (Looking in indexes: %s)" % ", ".join(index_urls)
    )

    with pytest.raises(env.PackageNotFound, match=re.escape(expected)):
        env.process_env.install_for_config(
            [inmanta.util.parse_requirement(requirement="this-package-does-not-exist")],
            config=PipConfig(
                index_url=index_urls[0],
                # The first element should only be passed to the index_url. If there are indexes in the environment
                # they should not be passed in the extra_index_url as they are already present in PIP_EXTRA_INDEX_URL
                # (second and third element of index_urls).
                extra_index_url=index_urls[3:] if (use_system_config and use_extra_indexes_env) else index_urls[1:],
                use_system_config=use_system_config,
            ),
        )


@pytest.mark.parametrize_any("use_system_config", [True, False])
def test_process_env_install_no_index(tmpdir: py.path.local, monkeypatch, use_system_config: bool, deactive_venv) -> None:
    """
    Attempt to install a package that does not exist with --no-index.
    To have --no-index set in the pip cmd, the config should not contain an index_url,
    we should not be using the system config and a path needs to be specified.
    it can also be set in the env_vars
    Assert the appropriate error is raised.
    """
    if use_system_config:
        monkeypatch.setenv("PIP_NO_INDEX", "true")

    setup_py_content = """
from setuptools import setup
setup(name="test")
"""
    # Write the minimal setup.py content to the temporary directory
    setup_py_path = os.path.join(tmpdir, "setup.py")
    with open(setup_py_path, "w") as setup_file:
        setup_file.write(setup_py_content)

    # We have two possible errors:
    # we install two packages, that fail differently
    # The order of failure is not fixed
    expected = re.escape("Packages this-package-does-not-exist were not found. No indexes were used.")
    other_expected = r"Packages setuptools.*were not found\. No indexes were used\."

    with pytest.raises(env.PackageNotFound, match=f"{expected}|{other_expected}"):
        deactive_venv.install_for_config(
            requirements=[inmanta.util.parse_requirement(requirement="this-package-does-not-exist")],
            paths=[env.LocalPackagePath(path=str(tmpdir))],
            config=PipConfig(use_system_config=use_system_config),
            add_inmanta_requires=False,
        )


@pytest.mark.slowtest
def test_process_env_install_from_index_conflicting_reqs(
    tmpdir: str, tmpvenv_active: tuple[py.path.local, py.path.local]
) -> None:
    """
    Attempt to install a package with conflicting version requirements from a pip index. Make sure this fails and the
    package remains uninstalled.
    """
    package_name: str = "more-itertools"
    with pytest.raises(env.ConflictingRequirements) as e:
        env.process_env.install_for_config(
            [inmanta.util.parse_requirement(requirement=f"{package_name}{version}") for version in [">8.5", "<=8"]],
            config=PipConfig(
                use_system_config=True,  # we need an upstream for some packages
            ),
        )
    assert "conflicting dependencies" in e.value.msg
    assert package_name not in env.process_env.get_installed_packages()


@pytest.mark.slowtest
@pytest.mark.parametrize("editable", [True, False])
def test_process_env_install_from_source(
    tmpvenv_active: tuple[py.path.local, py.path.local],
    modules_v2_dir: str,
    editable: bool,
) -> None:
    """
    Install a package from source into the process_env. Make sure the editable option actually results in an editable install.
    """
    package_name: str = "inmanta-module-minimalv2module"
    project_dir: str = os.path.join(modules_v2_dir, "minimalv2module")
    assert package_name not in env.process_env.get_installed_packages()
    env.process_env.install_from_source([env.LocalPackagePath(path=project_dir, editable=editable)])
    assert package_name in env.process_env.get_installed_packages()
    if editable:
        assert package_name in env.process_env.get_installed_packages(only_editable=True)


# v1 plugin loader overrides loader paths so verify that it doesn't interfere with env.process_env installs
@pytest.mark.parametrize("v1_plugin_loader", [True, False])
@pytest.mark.parametrize("package_name", ["lorem", "more-itertools", "inmanta-module-minimalv2module"])
@pytest.mark.slowtest
def test_active_env_get_module_file(
    local_module_package_index: str,
    tmpdir: py.path.local,
    tmpvenv_active: tuple[py.path.local, py.path.local],
    v1_plugin_loader: bool,
    package_name: str,
) -> None:
    """
    Test the env.ActiveEnv.get_module_file() command on a newly installed package. Make sure it works regardless of whether we
    install a dependency of inmanta-core (which is already installed in the encapsulating development venv), a new package or an
    inmanta module (namespace package).
    """
    venv_dir, _ = tmpvenv_active

    if package_name.startswith(module.ModuleV2.PKG_NAME_PREFIX):
        module_name = "inmanta_plugins." + package_name[len(module.ModuleV2.PKG_NAME_PREFIX) :].replace("-", "_")
        index = str(local_module_package_index)
        pip_config = PipConfig(
            index_url=index,
        )

    else:
        module_name = package_name.replace("-", "_")
        pip_config = PipConfig(
            use_system_config=True,
        )

    # unload module if already loaded from encapsulating development venv
    if module_name in sys.modules:
        loaded = [sub for sub in sys.modules.keys() if sub.startswith(module_name)]
        for sub in loaded:
            del sys.modules[sub]
    importlib.invalidate_caches()

    if v1_plugin_loader:
        loader.PluginModuleFinder.configure_module_finder([os.path.join(str(tmpdir), "libs")])

    assert env.ActiveEnv.get_module_file(module_name) is None
    env.process_env.install_for_config([inmanta.util.parse_requirement(requirement=package_name)], pip_config)
    assert package_name in env.process_env.get_installed_packages()
    module_info: Optional[tuple[Optional[str], Loader]] = env.ActiveEnv.get_module_file(module_name)
    assert module_info is not None
    module_file, mod_loader = module_info
    assert module_file is not None
    assert not isinstance(mod_loader, loader.PluginModuleLoader)
    assert module_file == os.path.join(env.process_env.site_packages_dir, *module_name.split("."), "__init__.py")
    # verify that the package was installed in the development venv
    assert str(venv_dir) in module_file
    importlib.import_module(module_name)
    assert module_name in sys.modules
    assert sys.modules[module_name].__file__ == module_file


@pytest.mark.slowtest
def test_active_env_get_module_file_editable_namespace_package(
    tmpdir: str,
    tmpvenv_active: tuple[py.path.local, py.path.local],
    modules_v2_dir: str,
    local_module_package_index,
) -> None:
    """
    Verify that get_module_file works after installing an editable namespace package in an active environment.
    """
    package_name: str = "inmanta-module-minimalv2module"
    module_name: str = "inmanta_plugins.minimalv2module"

    assert env.ActiveEnv.get_module_file(module_name) is None
    project_dir: str = os.path.join(modules_v2_dir, "minimalv2module")
    env.process_env.install_for_config(
        requirements=[],
        paths=[env.LocalPackagePath(path=project_dir, editable=True)],
        config=PipConfig(use_system_config=False, index_url=local_module_package_index),
        add_inmanta_requires=False,
    )
    assert package_name in env.process_env.get_installed_packages()
    module_info: Optional[tuple[Optional[str], Loader]] = env.ActiveEnv.get_module_file(module_name)
    assert module_info is not None
    module_file, mod_loader = module_info
    assert module_file is not None
    assert not isinstance(mod_loader, loader.PluginModuleLoader)
    assert module_file == os.path.join(modules_v2_dir, "minimalv2module", *module_name.split("."), "__init__.py")
    importlib.import_module(module_name)
    assert module_name in sys.modules
    assert sys.modules[module_name].__file__ == module_file
    # legacy method
    # We call it here to make sure the legacy compatibility works
    # It massages the inputs to get to the same call as the one above.
    # It should hit the cache there and return here.
    # Cheap and fast test
    env.process_env.install_from_source(
        paths=[env.LocalPackagePath(path=project_dir, editable=True)],
    )


def create_install_package(
    name: str, version: version.Version, requirements: list[inmanta.util.CanonicalRequirement], local_module_package_index: str
) -> None:
    """
    Creates and installs a simple package with specified requirements. Creates package in a temporary directory and
    cleans it up after install.

    :param name: Package name.
    :param version: Version for this package.
    :param requirements: Requirements on other packages. Required packages must already be installed when calling this function.
    :param local_module_package_index: upstream index to get setuptools and wheel
    """
    req_string: str = (
        "" if len(requirements) == 0 else ("[options]\ninstall_requires=" + "\n    ".join(str(req) for req in requirements))
    )
    with tempfile.TemporaryDirectory() as tmpdir:
        with open(os.path.join(tmpdir, "setup.cfg"), "w") as fd:
            fd.write(
                f"""
[metadata]
name = {name}
version = {version}

{req_string}
                """.strip()
            )
        with open(os.path.join(tmpdir, "pyproject.toml"), "w") as fd:
            fd.write(
                """
[build-system]
requires = ["setuptools", "wheel"]
build-backend = "setuptools.build_meta"
                """.strip()
            )
        env.process_env.install_for_config(
            requirements=[],
            paths=[env.LocalPackagePath(path=str(tmpdir), editable=False)],
            config=PipConfig(
                use_system_config=False,
                index_url=local_module_package_index,
            ),
        )


@pytest.mark.slowtest
def test_override_inmanta_package(tmpvenv_active_inherit: env.VirtualEnv) -> None:
    """
    Ensure that an ActiveEnv cannot override the main inmanta packages: inmanta-service-orchestrator, inmanta, inmanta-core.
    """
    installed_pkgs = tmpvenv_active_inherit.get_installed_packages()
    assert "inmanta-core" in installed_pkgs, "The inmanta-core package should be installed to run the tests"

    inmanta_requirements = inmanta.util.parse_requirement(requirement="inmanta-core==4.0.0")
    with pytest.raises(env.ConflictingRequirements) as excinfo:
        tmpvenv_active_inherit.install_for_config(
            requirements=[inmanta_requirements],
            config=PipConfig(
                use_system_config=True,  # we need some upstream
            ),
        )
    match = re.search(
        r"Cannot install (inmanta-core==4\.0\.0 and inmanta-core=.*|inmanta-core=.* and inmanta-core==4\.0\.0) because these "
        r"package versions have conflicting dependencies",
        excinfo.value.msg,
    )
    assert match is not None


@pytest.mark.slowtest
@pytest.mark.parametrize("invalid_char", ['"', "$", "`"])
def test_invalid_chars_in_venv_path(tmpdir, invalid_char: str) -> None:
    """
    Test that an error is raised when attempting to create a venv with invalid chars in its path.
    """
    venv_name = f"test{invalid_char}test"
    venv_dir = os.path.join(tmpdir, venv_name)

    with pytest.raises(ValueError) as excinfo:
        env.VirtualEnv(venv_dir)
    assert (
        f"Cannot create virtual environment because the provided path `{venv_dir}` contains an"
        f" invalid character (`{invalid_char}`)."
    ) in str(excinfo.value)


@pytest.mark.slowtest
def test_cache_on_active_env(tmpvenv_active_inherit: env.ActiveEnv, local_module_package_index: str) -> None:
    """
    Test whether the cache on an active env works correctly.
    """

    def _assert_install(requirement: str, installed: bool) -> None:
        parsed_requirement = inmanta.util.parse_requirement(requirement=requirement)
        for r in [requirement, parsed_requirement]:
            assert tmpvenv_active_inherit.are_installed(requirements=[r]) == installed

    _assert_install("inmanta-module-elaboratev2module==1.2.3", installed=False)
    tmpvenv_active_inherit.install_for_config(
        requirements=[inmanta.util.parse_requirement(requirement="inmanta-module-elaboratev2module==1.2.3")],
        config=PipConfig(
            index_url=local_module_package_index,
        ),
    )
    _assert_install("inmanta-module-elaboratev2module==1.2.3", installed=True)
    _assert_install("inmanta-module-elaboratev2module~=1.2.0", installed=True)
    _assert_install("inmanta-module-elaboratev2module<1.2.4", installed=True)
    _assert_install("inmanta-module-elaboratev2module>1.2.3", installed=False)
    _assert_install("inmanta-module-elaboratev2module==1.2.4", installed=False)


@pytest.mark.slowtest
def test_basic_logging(tmpdir, caplog):
    with caplog.at_level(logging.INFO):
        env_dir1 = tmpdir.mkdir("env1").strpath

        venv1 = env.VirtualEnv(env_dir1)

        venv1.use_virtual_env()

        log_sequence = LogSequence(caplog)
        log_sequence.assert_not("inmanta.env", logging.INFO, f"Creating new virtual environment in {env_dir1}")
        log_sequence.contains("inmanta.env", logging.INFO, f"Initializing virtual environment at {env_dir1}")


@pytest.mark.slowtest
def test_are_installed_dependency_cycle_on_extra(tmpdir, tmpvenv_active_inherit: env.VirtualEnv) -> None:
    """
    Ensure that the `ActiveEnv.are_installed()` method doesn't go into an infinite loop when there is a circular dependency
    involving an extra.

    Dependency loop:
        pkg[optional]
           -> dep[optional]
               -> pkg[optional]
    """
    pip_index = PipIndex(artifact_dir=str(tmpdir))
    create_python_package(
        name="pkg",
        pkg_version=version.Version("1.0.0"),
        path=os.path.join(tmpdir, "pkg"),
        publish_index=pip_index,
        optional_dependencies={
            "optional-pkg": [inmanta.util.parse_requirement(requirement="dep[optional-dep]")],
        },
    )
    create_python_package(
        name="dep",
        pkg_version=version.Version("1.0.0"),
        path=os.path.join(tmpdir, "dep"),
        publish_index=pip_index,
        optional_dependencies={
            "optional-dep": [inmanta.util.parse_requirement(requirement="pkg[optional-pkg]")],
        },
    )

    requirements = [inmanta.util.parse_requirement(requirement="pkg[optional-pkg]")]
    tmpvenv_active_inherit.install_for_config(
        requirements=requirements,
        config=PipConfig(
            index_url=pip_index.url,
        ),
    )
    assert tmpvenv_active_inherit.are_installed(requirements=requirements)


@pytest.mark.slowtest
def test_pip_logs(caplog, tmpvenv_active_inherit: str) -> None:
    """
    Verify the logs of a pip install:
        - all records start with 'inmanta.pip'
        - content of requirements and constraints files are logged
        - the pip command is logged
    """
    caplog.set_level(logging.DEBUG)

    with tempfile.TemporaryDirectory() as tmpdir:
        requirement1 = os.path.join(tmpdir, "requirement1.txt")
        requirement2 = os.path.join(tmpdir, "requirement2.txt")
        constraint1 = os.path.join(tmpdir, "constraint1.txt")
        constraint2 = os.path.join(tmpdir, "constraint2.txt")
        with open(requirement1, "w") as fd:
            fd.write(
                """
inmanta-module-std

                """
            )
        with open(requirement2, "w") as fd:
            fd.write(
                """
inmanta-module-net

inmanta-module-ip
                """
            )
        with open(constraint1, "w") as fd:
            fd.write(
                """
inmanta-module-std
                """
            )
        with open(constraint2, "w") as fd:
            fd.write(
                """

inmanta-module-ip
inmanta-module-net


                """
            )
        caplog.clear()
        Pip.run_pip_install_command_from_config(
            python_path=env.process_env.python_path,
            constraints_files=[constraint1, constraint2],
            requirements_files=[requirement1, requirement2],
            config=PipConfig(use_system_config=True),
        )

        assert all(record.name == "inmanta.pip" for record in caplog.records)
        python_path: str = tmpvenv_active_inherit.python_path
        assert (
            f"""
Content of requirements files:
    {requirement1}:
        inmanta-module-std
    {requirement2}:
        inmanta-module-net
        inmanta-module-ip
Content of constraints files:
    {constraint1}:
        inmanta-module-std
    {constraint2}:
        inmanta-module-ip
        inmanta-module-net
Pip command: {python_path} -m pip install -c {constraint1} -c {constraint2} -r {requirement1} -r {requirement2}
""".strip()
            in caplog.messages
        )



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    await connection.execute(
        """
CREATE TABLE IF NOT EXISTS public.resource_persistent_state (
    environment uuid NOT NULL REFERENCES public.environment(id) ON DELETE CASCADE,
    resource_id character varying NOT NULL,
    last_deploy timestamp with time zone,
    last_success timestamp with time zone,
    last_produced_events timestamp with time zone,
    last_deployed_attribute_hash character varying,
    last_deployed_version integer,
    last_non_deploying_status public.non_deploying_resource_state
        DEFAULT 'available'::public.non_deploying_resource_state NOT NULL,
    PRIMARY KEY(environment, resource_id)
);


-- The INSERT query uses a subquery to find the maximum model version
-- for each resource within each environment. This also includes orphans.
-- The query uses the 'resource_env_resourceid_index' index to efficiently find
-- the latest version for each resource within each environment without scanning the entire table.
INSERT INTO public.resource_persistent_state (
            environment,
            resource_id,
            last_deploy,
            last_success,
            last_non_deploying_status,
            last_produced_events,
            last_deployed_version,
            last_deployed_attribute_hash
        )
SELECT
    r.environment,
    r.resource_id,
    r.last_deploy,
    r.last_success,
    r.last_non_deploying_status,
    r.last_produced_events,
    r.model AS last_deployed_version,
    r.attribute_hash AS last_deployed_attribute_hash
FROM
    public.resource r
INNER JOIN (
    SELECT
        resource.environment,
        resource.resource_id,
        MAX(resource.model) AS max_model
    FROM
        public.resource
    JOIN public.configurationmodel ON
        public.resource.model = public.configurationmodel.version AND
        public.resource.environment = public.configurationmodel.environment
    WHERE
        public.configurationmodel.released = true
    GROUP BY
        resource.environment, resource.resource_id
) rmax ON
    r.environment = rmax.environment AND
    r.resource_id = rmax.resource_id AND
    r.model = rmax.max_model;

ALTER TABLE public.resource DROP COLUMN last_success;
ALTER TABLE public.resource DROP COLUMN last_non_deploying_status;
ALTER TABLE public.resource DROP COLUMN last_produced_events;
ALTER TABLE public.resource DROP COLUMN last_deploy;

"""
    )



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import functools
import typing
from collections.abc import Mapping, Set
from dataclasses import dataclass
from enum import IntEnum
from typing import Callable, Optional

from inmanta.deploy import tasks
from inmanta.types import ResourceIdStr

"""
Type alias for the union of all task types. Allows exhaustive case matches.
"""


class TaskPriority(IntEnum):
    # These priorities can be freely updated, we only care about the relative order
    TERMINATED = -1
    USER_DEPLOY = 0
    NEW_VERSION_DEPLOY = 1
    USER_REPAIR = 2
    DRYRUN = 3
    INTERVAL_DEPLOY = 4
    FACT_REFRESH = 5
    INTERVAL_REPAIR = 6


class PrioritizedTask[T: tasks.Task](typing.Protocol):
    """
    Resource action task with a priority attached. Lower values represent a higher priority. Negative priorities are reserved
    for internal use.

    Intended mostly for internal use within this module. External modules typically only care for MotivatedTask.

    :param task: The task
    :param priority: The priority assigned to the task
    :param requested_at: Integer representing when this task was requested with this priority. Serves as priority tiebreaker.
        Index is preserved for the lifetime of the task at the given priority, even if it is rescheduled due to changing
        dependencies. When the priority changes (is bumped), it is considered a "new request" (only from then on we desire to
        execute the task at that priority, it should not jump ahead of equal-priority tasks). This integer should be a unique,
        strictly monotonically rising (with time) value among all scheduled tasks.
    """

    task: T
    priority: TaskPriority
    requested_at: int


class MotivatedTask[T: tasks.Task](typing.Protocol):
    """
    Resource action task with an optional reason field attached.
    """

    task: T
    reason: Optional[str]


class TaskSpec[T: tasks.Task](PrioritizedTask[T], MotivatedTask[T], typing.Protocol):
    """
    Full specification of a resource action task request. Includes both priorities and reason.

    Intended mostly for internal use within this module. External modules typically only care for MotivatedTask.
    """

    task: T


@functools.total_ordering
@dataclass(kw_only=True)
class TaskQueueItem(TaskSpec[tasks.Task]):
    """
    Task item for the task queue. Tasks are prioritized based on TaskPriority, with insert order as tiebreaker.

    Implements full ordering based purely on priority and insert order (tiebreaker). Do not rely on == for task equality.

    The insert order field is a simple mechanic to maintain insert order for equal-priority tasks, as suggested in the
    heapq (collection type used by the agent queues) docs. The value should be monotonically rising within a queue. Gaps are
    allowed.
    """

    task: tasks.Task
    priority: TaskPriority
    requested_at: int
    reason: Optional[str]

    # Mutable state
    deleted: bool = False

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TaskQueueItem):
            return NotImplemented
        return (self.priority, self.requested_at, self.deleted) == (
            other.priority,
            other.requested_at,
            other.deleted,
        )

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, TaskQueueItem):
            return NotImplemented
        return (self.priority, self.requested_at, not self.deleted) < (
            other.priority,
            other.requested_at,
            not other.deleted,
        )


class AgentQueues:
    """
    Per-agent priority queue for ready-to-execute tasks. Clients must not interact with the underlying priority queues directly,
    only through the queue manipulation methods offered by this class.

    Each Task will be queued at most once. If an identical Task is queued later on, it will not be queued separately. It may
    however affect the current task's priority, causing it to be executed earlier than it would have otherwise.
    """

    # FIXME[#8019]: relies on undocumented asyncio.PriorityQueue._queue field and the fact that it's a heapq,
    #               can we do something about that?

    def __init__(self, new_agent_notify: Callable[[str], None]) -> None:
        """
        :param new_agent_notify: method to notify consumer about a new agent, called with agent name as argument.
        """
        self._new_agent_notify: Callable[[str], None] = new_agent_notify

        self._agent_queues: dict[str, asyncio.PriorityQueue[TaskQueueItem]] = {}
        # can not drop tasks from queue without breaking the heap invariant, or potentially breaking asyncio.Queue invariants
        # => take approach suggested in heapq docs: simply mark as deleted.
        # => Keep view on all active tasks for a given resource, which also doubles as lookup for client operations
        # Only tasks in this collection are considered queued, as far as this class' client interface is concerned.
        self._tasks_by_resource: dict[ResourceIdStr, dict[tasks.Task, TaskQueueItem]] = {}
        # monotonically rising value for item insert order
        # use simple counter rather than time.monotonic_ns() for performance reasons
        self._entry_count: int = 0
        self._in_progress: dict[tasks.Task, TaskPriority] = {}

    @property
    def in_progress(self) -> Mapping[tasks.Task, TaskPriority]:
        return self._in_progress

    def bump_in_progress_priorities(self, resources: Set[ResourceIdStr], priority: TaskPriority) -> None:
        """
        Bump the priority for currently running tasks. Affects priority of events sent out (and tasks created as a result)
        when one of these tasks finishes.
        """
        for resource in resources:
            task = tasks.Deploy(resource=resource)
            if task in self._in_progress and priority < self._in_progress[task]:
                self._in_progress[task] = priority

    def __contains__(self, item: tasks.Task) -> bool:
        """
        Return true iff the given task is queued in one of the agent queues.
        In progress tasks are not considered queued (anymore).
        """
        return item in self._tasks_by_resource.get(item.resource, {})

    def queued(self) -> Mapping[tasks.Task, TaskSpec[tasks.Task]]:
        """
        Returns all queued tasks and corresponding request details. Mostly intended for testing purposes.
        """
        return {t: spec for per_resource in self._tasks_by_resource.values() for t, spec in per_resource.items()}

    def reset(self) -> None:
        self._agent_queues.clear()
        self._tasks_by_resource.clear()
        self._entry_count = 0
        self._in_progress.clear()

    def _get_queue(self, agent_name: str) -> asyncio.PriorityQueue[TaskQueueItem]:
        """
        Return the queue for an agent, creating it if it does not exist.

        For internal use only. Queues must not be exposed to end user.
        """
        # All we do is sync and on the io loop, no need for locks!
        out = self._agent_queues.get(agent_name, None)
        if out is not None:
            return out
        out = asyncio.PriorityQueue()
        self._agent_queues[agent_name] = out
        self._new_agent_notify(agent_name)
        return out

    def get_tasks_for_resource(self, resource: ResourceIdStr) -> set[tasks.Task]:
        """
        Returns all queued tasks for a given resource id.
        """
        return set(self._tasks_by_resource.get(resource, {}).keys())

    def remove(self, task: tasks.Task) -> TaskSpec[tasks.Task]:
        """
        Removes the given task from its associated agent queue. Raises KeyError if it is not in the queue.
        Returns the priority at which the deleted task was queued.
        """
        the_tasks: dict[tasks.Task, TaskQueueItem] = self._tasks_by_resource.get(task.resource, {})
        queue_item: TaskQueueItem = the_tasks[task]
        queue_item.deleted = True
        del the_tasks[task]
        if not the_tasks:
            del self._tasks_by_resource[task.resource]
        return queue_item

    def discard(self, task: tasks.Task) -> Optional[TaskSpec[tasks.Task]]:
        """
        Removes the given task from its associated agent queue if it is present.
        Returns the priority at which the deleted task was queued, if it was at all.
        """
        try:
            return self.remove(task)
        except KeyError:
            return None

    def reserve_requested_at(self) -> int:
        """
        Reserve a unique, strictly monotonically rising value for a new task's requested_at value.
        """
        result: int = self._entry_count
        self._entry_count += 1
        return result

    def _queue_item_for_task(self, task: tasks.Task) -> Optional[TaskQueueItem]:
        return self._tasks_by_resource.get(task.resource, {}).get(task, None)

    def sorted(self, agent: str) -> list[MotivatedTask[tasks.Task]]:
        """
        Return a sorted view of the queued tasks for a single agent. Solely for testing purposes.
        """
        queue: asyncio.PriorityQueue[TaskQueueItem] = self._agent_queues[agent]
        backing_heapq: list[TaskQueueItem] = queue._queue  # type: ignore [attr-defined]
        backing_heapq.sort()
        return [item for item in backing_heapq if not item.deleted]

    ########################################
    # asyncio.PriorityQueue-like interface #
    ########################################

    def queue_put_nowait(
        self,
        task: tasks.Task,
        *,
        priority: TaskPriority,
        requested_at: Optional[int] = None,
        reason: Optional[str] = None,
    ) -> None:
        """
        Add a new task to the associated agent's queue.

        :param task: The task to queue.
        :param priority: The priority to queue the task at.
        :param requested_at: Integer value representing when this task was initially requested for the given priority (e.g. if
            it was deferred for some time. Serves as tiebreaker for comparing this task with equal-priority tasks. If none, the
            next value in line is used, resulting in the task being added at the end of its priority queue. A value can be
            reserved (for adding to the queue in the future) through the reserve_requested_at() method.
        """
        already_queued: Optional[TaskQueueItem] = self._queue_item_for_task(task)
        if already_queued is not None and already_queued.priority <= priority:
            # task is already queued with equal or higher priority
            return
        # reschedule with new priority, no need to explicitly remove, this is achieved by setting self._tasks_by_resource
        if already_queued is not None:
            already_queued.deleted = True
        item: TaskQueueItem = TaskQueueItem(
            task=task,
            priority=priority,
            requested_at=(
                self.reserve_requested_at()
                if (
                    # no requested_at given => use default
                    requested_at is None
                    # rescheduling at higher priority => assign fresh requested_at
                    or already_queued is not None
                )
                else requested_at
            ),
            reason=reason,
        )
        if task.resource not in self._tasks_by_resource:
            self._tasks_by_resource[task.resource] = {}
        self._tasks_by_resource[task.resource][task] = item
        self._get_queue(task.id.agent_name).put_nowait(item)

    def send_shutdown(self, reason: str = "Scheduler shutdown") -> None:
        """
        Wake up all workers after shutdown is signalled
        """
        poison_pill = TaskQueueItem(
            task=tasks.PoisonPill(resource=ResourceIdStr("system::Terminate[all,stop=True]")),
            priority=TaskPriority.TERMINATED,
            requested_at=-1,
            reason=reason,
        )
        for queue in self._agent_queues.values():
            queue.put_nowait(poison_pill)

    # Make sure to return MotivatedTask rather than full TaskSpec because priorities might change and we don't want to give
    # caller the wrong impression.
    async def queue_get(self, agent: str) -> MotivatedTask[tasks.Task]:
        """
        Consume a task from an agent's queue. If the queue is empty, blocks until a task becomes available.
        """
        queue: asyncio.PriorityQueue[TaskQueueItem] = self._agent_queues[agent]
        while True:
            item: TaskQueueItem = await queue.get()
            if item.deleted:
                # task was marked as removed, ignore it and consume the next item from the queue
                queue.task_done()
                continue
            # remove from the queue since it's been picked up
            self.discard(item.task)
            # add the item to _in_progress with the correct priority
            self._in_progress[item.task] = item.priority
            return item

    def task_done(self, agent: str, task: tasks.Task) -> None:
        """
        Indicate that a formerly enqueued task for a given agent is complete.

        Used by queue consumers. For each get() used to fetch a task, a subsequent call to task_done() tells the corresponding
        agent queue that the processing on the task is complete.
        """
        del self._in_progress[task]
        self._agent_queues[agent].task_done()


@dataclass(kw_only=True)
class BlockedDeploy(TaskSpec[tasks.Deploy]):
    """
    Deploy task that is blocked on one or more of its dependencies (subset of its requires relation).
    """

    task: tasks.Deploy
    blocked_on: set[ResourceIdStr]
    priority: TaskPriority
    requested_at: int
    reason: Optional[str]


class ScheduledWork:
    """
    Collection of deploy tasks that should be executed. Manages a priority queue of ready-to-execute tasks per agent
    (AgentQueues), as well as an unordered collection of waiting deploy tasks (across agents). Maintains the following
    invariants in between method calls:
    - Each task exists at most once in the scheduled work. It is however possible for an identical task to be executing.
    - A resource's requires relation is enforced within the set of scheduled work, i.e. for any two deploy tasks in the
        scheduled work, they will be processed in requires order. Only direct requires are considered. It is the responsibility
        of the scheduler to include in-between resources as scheduled work if/when transitive requires ordering is desired.

    Expects to be informed by the scheduler of deploy requests and/or state changes through deploy_with_context() and
    delete_resource().

    Expects to be informed by scheduler of finished tasks through finished_deploy().

    :param requires: Live, read-only view on requires mapping for the latest model state.
    :param provides: Live, read-only view on provides mapping for the latest model state.
    :param new_agent_notify: Method to notify client of newly created agent queues. When notified about a queue, client is
        expected to start consuming its tasks.
    """

    def __init__(
        self,
        requires: Mapping[ResourceIdStr, Set[ResourceIdStr]],
        provides: Mapping[ResourceIdStr, Set[ResourceIdStr]],
        new_agent_notify: Callable[[str], None],
    ) -> None:
        """
        :param requires: This is a view on the requires of the requires field from the associated ModelState object.
                         It's a view, which means that it's updated whenever the original RequiresProvides mapping is updated.
        :param provides: This is a view on the provides of the requires field from the associated ModelState object.
                         It's a view, which means that it's updated whenever the original RequiresProvides mapping is updated.
        """
        self.requires: Mapping[ResourceIdStr, Set[ResourceIdStr]] = requires
        self.provides: Mapping[ResourceIdStr, Set[ResourceIdStr]] = provides
        self.agent_queues: AgentQueues = AgentQueues(new_agent_notify)
        self._waiting: dict[ResourceIdStr, BlockedDeploy] = {}

    def reset(self) -> None:
        self.agent_queues.reset()
        self._waiting.clear()

    def add_poison_pill_to_agent_queues(self, reason: str) -> None:
        self.agent_queues.send_shutdown(reason=reason)

    def deploy_with_context(
        self,
        resources: Set[ResourceIdStr],
        reason: str,
        *,
        priority: TaskPriority,
        deploying: Optional[Set[ResourceIdStr]] = None,
        added_requires: Optional[Mapping[ResourceIdStr, Set[ResourceIdStr]]] = None,
        dropped_requires: Optional[Mapping[ResourceIdStr, Set[ResourceIdStr]]] = None,
        force_deploy: bool = False,
    ) -> None:
        """
        Add deploy tasks for the given resources. Additionally update the scheduled work state to reflect the new model state
        and scheduler intent. May defer tasks that were previously considered ready to execute if any of their dependencies are
        added to the scheduled work or if new dependencies are added.

        :param resources: Set of resources that should be deployed. Adds a deploy task to the scheduled work for each
            of these, unless it is already scheduled.
        :param reason: The reason for this deploy.
        :param priority: The priority of this deploy.
        :param deploying: Set of resources for which a non-stale deploy is in progress, i.e. the scheduler does not need to
            take action to deploy the latest intent for any of these resources because that deploy is already in progress
            (it will still ensure they are scheduled if they have gotten new dependencies).
        :param added_requires: Requires edges that were added since the previous state update, if any.
        :param dropped_requires: Requires edges that were removed since the previous state update, if any.
        :param force_deploy: Force the deploy of the given resources, even if a deploy is already running for the same intent.
        """
        deploying = deploying if deploying is not None else set()
        added_requires = added_requires if added_requires is not None else {}
        dropped_requires = dropped_requires if dropped_requires is not None else {}

        # set of resources for which the blocked_on field was updated without checking whether it has become unblocked
        maybe_runnable: set[ResourceIdStr] = set()

        # lookup caches for visited nodes
        # queued or running, pre-populate with in-progress deploys
        queued: set[ResourceIdStr] = set(deploying) if not force_deploy else set(deploying - resources)
        not_scheduled: set[ResourceIdStr] = set()

        # Bump the priority of (non-stale) deploying tasks. This will only increase the priority of propagated events
        self.agent_queues.bump_in_progress_priorities(resources & deploying, priority)

        # First drop all dropped requires so that we work on the smallest possible set for this operation.
        for resource, dropped in dropped_requires.items():
            if not dropped:
                # empty set, nothing to do
                continue
            if resource not in self._waiting:
                # this resource is not currently waiting for anything => nothing to do
                continue
            self._waiting[resource].blocked_on.difference_update(dropped)
            maybe_runnable.add(resource)

        def is_scheduled(resource: ResourceIdStr) -> bool:
            """
            Returns whether the resource is currently scheduled, caching the results.
            """
            # start with cheap checks: check waiting and cached sets
            if resource in self._waiting or resource in queued:
                # definitely scheduled
                return True
            if resource in not_scheduled:
                # definitely not scheduled
                return False
            # finally, check more expensive agent queue
            task: tasks.Deploy = tasks.Deploy(resource=resource)
            if task in self.agent_queues:
                # populate cache
                queued.add(resource)
                return True
            # populate cache
            not_scheduled.add(resource)
            return False

        def extend_blocked_on(resource: ResourceIdStr, new_blockers: set[ResourceIdStr]) -> None:
            """
            Add to the blocked on set for this resource, deferring the associated deploy task if it is already queued.
            This method takes ownership of the new_blockers set, the caller should not use it again.

            :param resource: The resource for which to add blockers.
            :param new_blockers: The resources to add as blockers. This must be a subset of the resources's scheduled requires,
                i.e. requires that are currently queued or waiting. The method takes ownership of this object.
            """
            if not new_blockers:
                # empty set, nothing to do
                return
            if not is_scheduled(resource):
                # resource is not scheduled => nothing to do
                return
            if resource in queued:
                # we're adding a dependency so it's definitely not ready to execute anymore
                # move from agent queue to waiting
                #
                # discard rather than remove because task may already be running, in which case we leave it run its course
                # and simply add a new one
                task: tasks.Deploy = tasks.Deploy(resource=resource)
                discarded: Optional[TaskSpec[tasks.Task]] = self.agent_queues.discard(task)

                new_priority: TaskPriority
                new_requested_at: int
                new_reason: Optional[str]
                # always use the previously scheduled priority, because new priority only applies to anything in 'resources',
                # which are scheduled seperately.
                if discarded is not None:
                    new_priority, new_requested_at, new_reason = (
                        discarded.priority,
                        discarded.requested_at,
                        discarded.reason,
                    )
                else:
                    # Get priority from running task. This should always return something, but default to this request's
                    # priority just in case
                    new_priority = self.agent_queues.in_progress.get(task, priority)
                    # old task is already running, consider this a new request
                    new_requested_at = self.agent_queues.reserve_requested_at()
                    new_reason = "rescheduling because a dependency was scheduled while it was deploying"

                queued.remove(resource)
                self._waiting[resource] = BlockedDeploy(
                    task=task,
                    # task was previously ready to execute => assume no other blockers than this one
                    blocked_on=new_blockers,
                    priority=new_priority,
                    requested_at=new_requested_at,
                    reason=new_reason,
                )
            else:
                self._waiting[resource].blocked_on.update(new_blockers)
                maybe_runnable.discard(resource)

        # ensure desired resource deploys are scheduled
        for resource in resources:
            task = tasks.Deploy(resource=resource)
            if is_scheduled(resource):
                # Deploy is already scheduled / running. Check to see if this task has a higher priority than the one already
                # scheduled. If it has, update the priority. If any of its dependencies are to be
                # scheduled as well, they will follow the provides relation to ensure this deploy waits its turn.
                if task in self.agent_queues:
                    # simply add it again, the queue will make sure only the highest priority is kept
                    self.agent_queues.queue_put_nowait(task, priority=priority, reason=reason)
                elif resource in self._waiting and priority < self._waiting[resource].priority:
                    self._waiting[resource].priority = priority
                    # bumping priority => consider this a new request, resetting the requested_at field
                    # (see TaskPriority.requested_at docstring)
                    self._waiting[resource].requested_at = self.agent_queues.reserve_requested_at()
                    self._waiting[resource].reason = reason
                continue
            # task is not yet scheduled, schedule it now
            blocked_on: set[ResourceIdStr] = {
                dependency for dependency in self.requires.get(resource, ()) if is_scheduled(dependency)
            }
            self._waiting[resource] = BlockedDeploy(
                task=tasks.Deploy(resource=resource),
                blocked_on=blocked_on,
                priority=priority,
                requested_at=self.agent_queues.reserve_requested_at(),
                reason=reason,
            )
            not_scheduled.discard(resource)
            if not blocked_on:
                # not currently blocked on anything but new dependencies may still be scheduled => mark for later check
                maybe_runnable.add(resource)

            # inform along provides relation that this task has been scheduled, deferring already scheduled provides
            for dependant in self.provides.get(resource, ()):
                extend_blocked_on(dependant, {resource})

        # update state for added requires
        for resource, new in added_requires.items():
            extend_blocked_on(resource, {r for r in new if is_scheduled(r)})

        # finally check if any tasks have become ready to run
        for resource in maybe_runnable:
            blocked: BlockedDeploy = self._waiting[resource]
            self._queue_if_ready(blocked)
            # no more need to update cache entries

    def _queue_if_ready(self, blocked_deploy: BlockedDeploy) -> None:
        """
        Check if the given deploy has become unblocked and move to the agent queues if it has.
        """
        if blocked_deploy.blocked_on:
            # still waiting for something, nothing to do
            return
        # ready to execute, move to agent queue
        self.agent_queues.queue_put_nowait(
            blocked_deploy.task,
            priority=blocked_deploy.priority,
            requested_at=blocked_deploy.requested_at,
            reason=blocked_deploy.reason,
        )
        del self._waiting[blocked_deploy.task.resource]

    def delete_resource(self, resource: ResourceIdStr) -> None:
        """
        Drop tasks for a given resource when it was deleted from the model or
        when we know it can't progress e.g. it is undefined and thus known to be blocked
        on another resource.
        Does not affect dry-run tasks because they do not act on the latest desired state.

        Does nothing if no work is found for this resource.
        """
        # delete from waiting collection if deploy task is waiting to be queued
        if resource in self._waiting:
            del self._waiting[resource]
        # additionally delete from agent_queues if a task is already queued
        task: tasks.Task
        for task in self.agent_queues.get_tasks_for_resource(resource):
            if task.delete_with_resource():
                self.agent_queues.discard(task)

    def finished_deploy(self, resource: ResourceIdStr) -> None:
        """
        Report that a resource has finished deploying, regardless of the intent version (including stale deploys)
        and regardless of the deploy result (success / failure).
        """
        if resource in self._waiting or tasks.Deploy(resource=resource) in self.agent_queues:
            # A new deploy task was scheduled in the meantime, no need to do anything else
            # This check also happens to cover the case of stale deploys. They are always expected to have a non-stale
            # one in the queue, so they would fall in this NOOP condition. If for some reason (e.g. due to a bug in the
            # scheduler) this invariant is not met, we'd rather clean up the state that have the risk of stuck deploys.
            # This is why the docstring states that all deploys, even stale ones, should be reported, even if we expect
            # stale ones to be a NOOP.
            return
        for dependant in self.provides.get(resource, []):
            blocked_deploy: Optional[BlockedDeploy] = self._waiting.get(dependant, None)
            if blocked_deploy is None:
                # dependant is not currently scheduled
                continue
            # remove the finished resource from the blocked on set and check if that unblocks the dependant
            blocked_deploy.blocked_on.discard(resource)
            self._queue_if_ready(blocked_deploy)



"""
Copyright 2023 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import logging
import os.path
import sys
import uuid
from asyncio import TimeoutError, subprocess, wait_for
from collections.abc import Mapping
from io import StringIO
from typing import Optional

import pytest
import yaml

import inmanta
from inmanta import config
from inmanta.config import logging_config
from inmanta.const import ENVIRON_FORCE_TTY
from inmanta.logging import InmantaLoggerConfig, LoggingConfigBuilder, LoggingConfigFromFile, Options


def load_config_file_to_dict(file_name: str, context: Mapping[str, str]) -> dict[str, object]:
    logging_config_source = LoggingConfigFromFile(file_name=file_name)
    return logging_config_source.read_logging_config(context=context)


@pytest.fixture(autouse=True)
def cleanup_logger():
    root_log_level = logging.root.level
    InmantaLoggerConfig.clean_instance()
    yield
    # Make sure we maintain the initial root log level, so that logging in pytest works as expected.
    logging.root.setLevel(root_log_level)


@pytest.mark.parametrize_any(
    "log_file, log_file_level, verbose",
    [(None, "INFO", 1), (None, "ERROR", 4), ("test.log", "WARNING", 4), ("test.log", "DEBUG", 4)],
)
def test_apply_options(tmpdir, log_file, log_file_level, verbose, allow_overriding_root_log_level: None):
    stream = StringIO()
    inmanta_logger = InmantaLoggerConfig.get_instance(stream)
    logger = logging.getLogger("test_logger")

    if log_file:
        log_file = tmpdir.join("test.log")

    options1 = Options(log_file=log_file, log_file_level=log_file_level, verbose=verbose)
    inmanta_logger.apply_options(options1)
    logger.debug("debug: This is the first test")
    logger.info("info: This is the second test")
    logger.warning("warning: This is the third test")
    logger.error("error: This is the fourth test")
    if not log_file:
        log_output = stream.getvalue().strip()
        debug_in_output = "test_logger              DEBUG   debug: This is the first test" in log_output
        info_in_output = "test_logger              INFO    info: This is the second test" in log_output
        warning_in_output = "test_logger              WARNING warning: This is the third test" in log_output
        error_in_output = "test_logger              ERROR   error: This is the fourth test" in log_output
        assert debug_in_output if int(verbose) >= 3 else not debug_in_output
        assert info_in_output if int(verbose) >= 2 else not info_in_output
        assert warning_in_output if int(verbose) >= 1 else not warning_in_output
        assert error_in_output

    else:
        with open(str(log_file)) as f:
            log_output = f.read().strip()
            debug_in_output = "DEBUG    test_logger debug: This is the first test" in log_output
            info_in_output = "INFO     test_logger info: This is the second test" in log_output
            warning_in_output = "WARNING  test_logger warning: This is the third test" in log_output
            error_in_output = "ERROR    test_logger error: This is the fourth test" in log_output
            assert debug_in_output if log_file_level in ["DEBUG"] else not debug_in_output
            assert info_in_output if log_file_level in ["DEBUG", "INFO"] else not info_in_output
            assert warning_in_output if log_file_level in ["WARNING", "INFO", "DEBUG"] else not warning_in_output
            assert error_in_output


def test_logging_apply_options_2_times(allow_overriding_root_log_level: None):
    stream = StringIO()
    inmanta_logger = InmantaLoggerConfig.get_instance(stream)
    options1 = Options(log_file=None, log_file_level="INFO", verbose="1")
    inmanta_logger.apply_options(options1)
    with pytest.raises(Exception) as e:
        options2 = Options(log_file=None, log_file_level="INFO", verbose="2")
        inmanta_logger.apply_options(options2)
    message = "Options can only be applied once to a handler."
    assert message in str(e.value)


def test_logging_cleaned_after_apply_options(tmpdir, allow_overriding_root_log_level: None):
    # verifies that when changing the stream with apply_option, the old stream is properly cleaned up
    # and not used anymore.
    stream = StringIO()
    inmanta_logger = InmantaLoggerConfig.get_instance(stream)
    logger = logging.getLogger("test_logger")

    logger.info("This is a test message")
    log_output = stream.getvalue().strip()
    expected_output = "test_logger              INFO    This is a test message"
    assert log_output == expected_output

    log_file = tmpdir.join("test.log")

    options3 = Options(log_file=log_file, log_file_level="WARNING", verbose="4")
    inmanta_logger.apply_options(options3)
    logger.warning("warning: This is the second test")
    with open(str(log_file)) as f:
        contents = f.read()
        assert "WARNING  test_logger warning: This is the second test" in contents

    log_output = stream.getvalue().strip()
    expected_output = "test_logger              INFO    This is a test message"
    assert log_output == expected_output


def test_handling_logging_config_option(tmpdir, monkeypatch, allow_overriding_root_log_level: None) -> None:
    """
    Verify the behavior of the logging_config option.
    """
    logger = logging.getLogger("TEST")

    stream = StringIO()
    # In order to reference an object in the logging_config file, it needs to be part of a module.
    # For this reason we add the 'pytest_stream' attribute to the inmanta module. It's referenced in the
    # logging_config file using ext://inmanta.pytest_stream
    monkeypatch.setattr(inmanta, "pytest_stream", stream, raising=False)

    def write_logging_config_file(path: str, formatter: str) -> None:
        config = {
            "version": 1,
            "formatters": {
                "console_formatter": {
                    "format": formatter,
                }
            },
            "handlers": {
                "console_handler": {
                    "class": "logging.StreamHandler",
                    "formatter": "console_formatter",
                    "level": "INFO",
                    "stream": "ext://inmanta.pytest_stream",
                },
            },
            "root": {
                "level": "INFO",
                "handlers": ["console_handler"],
            },
            "disable_existing_loggers": False,
        }
        with open(path, "w") as fh:
            yaml.dump(config, fh)

    def setup_logging_config(cli_options: Options, file_option_value: Optional[str] = None) -> None:
        # Set/Reset config options
        config.Config._reset()
        if file_option_value:
            config.logging_config.set(file_option_value)
        # Reset/Configure logging framework
        InmantaLoggerConfig.clean_instance()
        inmanta_logger_config = InmantaLoggerConfig.get_instance()
        inmanta_logger_config.apply_options(cli_options)
        # Reset stream buffer
        stream.truncate()

    def get_output_and_clear() -> str:
        value = stream.getvalue()
        stream.seek(0)
        stream.truncate(0)
        return value

    # Log file to pass in the options.
    # This config option will be overridden by the other config in this test case.
    old_log_file = os.path.join(tmpdir, "old_log_file.txt")
    path_logging_config_file1 = os.path.join(tmpdir, "logging_config1.yml")
    path_logging_config_file2 = os.path.join(tmpdir, "logging_config2.yml")
    path_logging_config_file3 = os.path.join(tmpdir, "logging_config3.yml")

    # Set --logging-config option on CLI only
    write_logging_config_file(path=path_logging_config_file1, formatter="AAA %(message)s")
    # Also assert that the other config options are ignored when logging_config is set.
    options1 = {"log_file": old_log_file, "log_file_level": "4", "keep_logger_names": True}
    setup_logging_config(
        cli_options=Options(logging_config=path_logging_config_file1, **options1),
        file_option_value=None,
    )
    logger.info("test")
    output = get_output_and_clear()
    assert "AAA test" in output
    assert (
        f"Ignoring the following options: --log-file {old_log_file}, --log-file-level 4, --keep-logger-names. "
        f"Using logging config from file {path_logging_config_file1}" in output
    )

    # Set logging_config option in cfg file only
    write_logging_config_file(path=path_logging_config_file1, formatter="BBB %(message)s")
    options2 = {"keep_logger_names": True}
    setup_logging_config(
        cli_options=Options(**options2),
        file_option_value=path_logging_config_file1,
    )
    logger.info("test")
    output = get_output_and_clear()
    assert "BBB test" in output
    assert (
        f"Ignoring the following options: --keep-logger-names. "
        f"Using logging config from file {path_logging_config_file1}" in output
    )

    # Set the logging-config config option both on CLI and cfg file. CLI option takes precedence.
    write_logging_config_file(path=path_logging_config_file1, formatter="CCC %(message)s")
    write_logging_config_file(path=path_logging_config_file2, formatter="DDD %(message)s")
    setup_logging_config(
        cli_options=Options(logging_config=path_logging_config_file1, **options2),
        file_option_value=path_logging_config_file2,
    )
    logger.info("test")
    output = get_output_and_clear()
    assert "CCC test" in output
    assert (
        f"Ignoring the following options: --keep-logger-names. "
        f"Using logging config from file {path_logging_config_file1}" in output
    )

    # Set the logging-config config option in the cfg config file and using the environment variable.
    # The environment variable takes precedence.
    write_logging_config_file(path=path_logging_config_file1, formatter="EEE %(message)s")
    write_logging_config_file(path=path_logging_config_file2, formatter="FFF %(message)s")
    options3 = {"timed": True}
    with monkeypatch.context() as m:
        m.setenv("INMANTA_CONFIG_LOGGING_CONFIG", path_logging_config_file1)
        setup_logging_config(
            cli_options=Options(**options3),
            file_option_value=path_logging_config_file2,
        )
        logger.info("test")
        output = get_output_and_clear()
        assert "EEE test" in output
        assert (
            f"Ignoring the following options: --timed-logs. "
            f"Using logging config from file {path_logging_config_file1}" in output
        )

    # Set the logging-config config option on the CLI, in the cfg config file and using the environment variable.
    # The CLI option takes precedence.
    write_logging_config_file(path=path_logging_config_file1, formatter="GGG %(message)s")
    write_logging_config_file(path=path_logging_config_file2, formatter="HHH %(message)s")
    write_logging_config_file(path=path_logging_config_file3, formatter="III %(message)s")
    options4 = {"log_file": old_log_file, **options3}
    with monkeypatch.context() as m:
        m.setenv("INMANTA_CONFIG_LOGGING_CONFIG", path_logging_config_file1)
        setup_logging_config(
            cli_options=Options(logging_config=path_logging_config_file2, **options4),
            file_option_value=path_logging_config_file3,
        )
        logger.info("test")
        output = get_output_and_clear()
        assert "HHH test" in output
        assert (
            f"Ignoring the following options: --log-file {old_log_file}, --timed-logs. "
            f"Using logging config from file {path_logging_config_file2}" in output
        )

    # After all of these logs, we assert that the log file passed to the options (old_log_file)
    # Was ignored and nothing was written to it.
    assert not os.path.exists(old_log_file)

    # If given no other options, we can still use the old logging cli options
    setup_logging_config(
        cli_options=Options(**options1),
    )
    logger.info("test")
    # Assert that the file gets created
    assert os.path.exists(old_log_file)
    with open(old_log_file, "r") as file:
        content = file.read()
        assert "test" in content
        assert "WARNING" not in content


def test_log_file_or_template(tmp_path):

    with pytest.raises(Exception, match=f"Logging config file {str(tmp_path / 'test')} doesn't exist."):
        # TODO: do we want more specific exceptions?
        load_config_file_to_dict(str(tmp_path / "test"), {})

    content_1 = {"test": "x"}
    content_2 = {"test": "{xvar}", "flah": "\n\n\n{yvar}\n", "{test}": "value"}

    f1 = tmp_path / "test.yaml"
    f2 = tmp_path / "test.yaml.tmpl"

    with open(f1, "w") as fh:
        yaml.dump(content_1, fh)

    with open(f2, "w") as fh:
        yaml.dump(content_2, fh)

    assert load_config_file_to_dict(str(f1), {}) == content_1

    with pytest.raises(
        Exception,
        match="The logging configuration template from .* refers to context variable 'test',"
        " but this variable is not available. The context is limited to xvar, yvar",
    ):
        load_config_file_to_dict(
            str(f2),
            {
                "xvar": "A",
                "yvar": "B",
            },
        )

    config = load_config_file_to_dict(str(f2), {"xvar": "A", "yvar": "B", "test": "key"})
    assert config == {"test": "A", "flah": "\n\n\nB\n", "key": "value"}

    # we control the values, so not very relevant from security perspective
    # overwrite type of injection
    config = load_config_file_to_dict(str(f2), {"xvar": "A", "yvar": "B", "test": "flah"})
    assert config == {"test": "A", "flah": "value"}

    # Full on injection
    config = load_config_file_to_dict(str(f2), {"xvar": "A", "yvar": "B", "test": "flah': 'zxxx'\ntest: zzz\nflah: zzz\n#"})
    assert config == {"test": "zzz", "flah": "zzz"}


def test_scheduler_documentation_conformance(inmanta_config, monkeypatch):
    monkeypatch.setenv(ENVIRON_FORCE_TTY, "yes")
    env = uuid.uuid4()
    from_file_dict = load_config_file_to_dict(
        os.path.join(os.path.dirname(__file__), "..", "misc/scheduler_log.yml.tmpl"), context={"environment": env}
    )
    default = LoggingConfigBuilder()
    from_config = default.get_logging_config_from_options(
        sys.stdout, Options(log_file_level="DEBUG"), component="scheduler", context={"environment": str(env)}
    )

    assert from_config._to_dict_config() == from_file_dict


def test_server_documentation_conformance(inmanta_config, monkeypatch):
    monkeypatch.setenv(ENVIRON_FORCE_TTY, "yes")
    from_file_dict = load_config_file_to_dict(os.path.join(os.path.dirname(__file__), "..", "misc/server_log.yml"), context={})

    default = LoggingConfigBuilder()
    from_config = default.get_logging_config_from_options(
        sys.stdout,
        Options(log_file_level="INFO", log_file="/var/log/inmanta/server.log", timed=True),
        component="server",
        context={},
    )

    assert from_config._to_dict_config() == from_file_dict


def test_logging_config_content_environment_variables(monkeypatch, capsys, tmpdir) -> None:
    """
    Verify that the environment variables, that contain the content of the logging configuration,
    are correctly taken into account when loading the logging configuration.
    """
    logging_config_file = os.path.join(tmpdir, "config.yml")
    with open(logging_config_file, "w") as fh:
        fh.write(
            """
                disable_existing_loggers: false
                formatters:
                  console_formatter:
                    format: "DONT_USE -- %(message)s"
                handlers:
                  console_handler:
                    class: logging.StreamHandler
                    formatter: console_formatter
                    level: INFO
                    stream: ext://sys.stdout
                root:
                  handlers:
                  - console_handler
                  level: INFO
                version: 1
            """
        )
    logging_config.set(logging_config_file)

    # Set the INMANTA_CONFIG_LOGGING_CONFIG_TMPL environment variable and verify that it overrides
    # the logging config set via the config.logging_config configuration option.
    logging_config1 = """
        disable_existing_loggers: false
        formatters:
          console_formatter:
            format: "config1 -- {environment} -- %(message)s"
        handlers:
          console_handler:
            class: logging.StreamHandler
            formatter: console_formatter
            level: INFO
            stream: ext://sys.stdout
        root:
          handlers:
          - console_handler
          level: INFO
        version: 1
    """
    env_id = str(uuid.uuid4())
    monkeypatch.setenv("INMANTA_CONFIG_LOGGING_CONFIG_TMPL", logging_config1)
    inmanta_logging_config = InmantaLoggerConfig.get_instance(stream=sys.stdout)
    inmanta_logging_config.apply_options(options=Options(), component="server", context={"environment": env_id})
    logger = logging.getLogger("test")
    capsys.readouterr()  # Clear buffer
    logger.info("test")
    captured = capsys.readouterr()
    assert f"config1 -- {env_id} -- test" in captured.out

    # Set the component-specific template and verify that it overrides the config from INMANTA_CONFIG_LOGGING_CONFIG_TMPL
    logging_config2 = """
        disable_existing_loggers: false
        formatters:
          console_formatter:
            format: "config2 -- {environment} -- %(message)s"
        handlers:
          console_handler:
            class: logging.StreamHandler
            formatter: console_formatter
            level: INFO
            stream: ext://sys.stdout
        root:
          handlers:
          - console_handler
          level: INFO
        version: 1
    """
    monkeypatch.setenv("INMANTA_LOGGING_SERVER_TMPL", logging_config2)
    inmanta_logging_config.clean_instance()
    inmanta_logging_config = InmantaLoggerConfig.get_instance(stream=sys.stdout)
    inmanta_logging_config.apply_options(options=Options(), component="server", context={"environment": env_id})
    logger = logging.getLogger("test")
    capsys.readouterr()  # Clear buffer
    logger.info("test")
    captured = capsys.readouterr()
    assert f"config2 -- {env_id} -- test" in captured.out

    # Set INMANTA_LOGGING_SERVER_TMPL AND INMANTA_LOGGING_SERVER_CONTENT simultaneously.
    # Assert that INMANTA_LOGGING_SERVER_CONTENT is used.
    logging_config3 = """
        disable_existing_loggers: false
        formatters:
          console_formatter:
            format: "config3 -- {environment} -- %(message)s"
        handlers:
          console_handler:
            class: logging.StreamHandler
            formatter: console_formatter
            level: INFO
            stream: ext://sys.stdout
        root:
          handlers:
          - console_handler
          level: INFO
        version: 1
    """
    monkeypatch.setenv("INMANTA_LOGGING_SERVER_CONTENT", logging_config3)
    inmanta_logging_config.clean_instance()
    inmanta_logging_config = InmantaLoggerConfig.get_instance(stream=sys.stdout)
    inmanta_logging_config.apply_options(options=Options(), component="server", context={"environment": env_id})
    captured = capsys.readouterr()
    assert (
        "Environment variables INMANTA_LOGGING_SERVER_CONTENT and INMANTA_LOGGING_SERVER_TMPL are set simultaneously."
        " Using INMANTA_LOGGING_SERVER_CONTENT" in captured.out
    )
    logger = logging.getLogger("test")
    logger.info("test")
    captured = capsys.readouterr()
    assert "config3 -- {environment} -- test" in captured.out

    # Verify that the --logging-config CLI option still overrides all other config.
    other_logging_config_file = os.path.join(tmpdir, "cli.yml")
    with open(other_logging_config_file, "w") as fh:
        fh.write(
            """
                disable_existing_loggers: false
                formatters:
                  console_formatter:
                    format: "CLI -- %(message)s"
                handlers:
                  console_handler:
                    class: logging.StreamHandler
                    formatter: console_formatter
                    level: INFO
                    stream: ext://sys.stdout
                root:
                  handlers:
                  - console_handler
                  level: INFO
                version: 1
            """
        )
    inmanta_logging_config.clean_instance()
    inmanta_logging_config = InmantaLoggerConfig.get_instance(stream=sys.stdout)
    inmanta_logging_config.apply_options(
        options=Options(logging_config=other_logging_config_file), component="server", context={"environment": env_id}
    )
    logger = logging.getLogger("test")
    capsys.readouterr()  # Clear buffer
    logger.info("test")
    captured = capsys.readouterr()
    assert "CLI -- test" in captured.out


async def test_output_default_logging_cmd(inmanta_config, tmp_path):

    components = ["scheduler", "server", "compiler"]
    for component in components:
        if component == "scheduler":
            output_file = str(tmp_path / f"{component}.yml.tmpl")
        else:
            output_file = str(tmp_path / f"{component}.yml")

        args = [
            sys.executable,
            "-m",
            "inmanta.app",
            "output-default-logging-config",
            "--component",
            component,
            output_file,
        ]
        process = await subprocess.create_subprocess_exec(*args, stdout=subprocess.PIPE)
        try:
            await wait_for(process.communicate(), timeout=5)
        except TimeoutError as e:
            process.kill()
            await process.communicate()
            raise e
        assert process.returncode == 0

        with open(output_file, "r") as fh:
            logging_config = fh.read()
        assert "no_color: false" in logging_config
        assert "reset: true" in logging_config
        assert "log_colors: null" not in logging_config

        # Assert we get an error if the output file already exists
        process = await subprocess.create_subprocess_exec(*args, stderr=subprocess.PIPE)
        try:
            (_, stderr) = await wait_for(process.communicate(), timeout=5)
            assert f"The requested output location already exists: {output_file}" in stderr.decode()
        except TimeoutError as e:
            process.kill()
            await process.communicate()
            raise e
        assert process.returncode == 1

        os.remove(output_file)

    # Assert that the .tmpl suffix is enforced when generating the logging config for the scheduler.
    output_file = str(tmp_path / "test.yml")
    args = [
        sys.executable,
        "-m",
        "inmanta.app",
        "output-default-logging-config",
        "--component",
        "scheduler",
        "-e",
        str(uuid.uuid4()),
        output_file,
    ]
    process = await subprocess.create_subprocess_exec(*args, stderr=subprocess.PIPE)
    try:
        (_, stderr) = await wait_for(process.communicate(), timeout=5)
    except TimeoutError as e:
        process.kill()
        await process.communicate()
        raise e
    assert process.returncode == 1
    assert (
        "The config being generated will be a template, but the given filename doesn't end with the .tmpl suffix."
        in stderr.decode()
    )
